# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_ide CLI command handlers."""

from glob import iglob
import logging
from pathlib import Path
import shlex
import shutil
import subprocess
import sys
from typing import cast, Set

from pw_cli.diff import print_diff
from pw_cli.env import pigweed_environment
from pw_cli.status_reporter import LoggingStatusReporter, StatusReporter

from pw_ide.activate import pigweed_root

from pw_ide.cpp import (
    COMPDB_FILE_NAME,
    ClangdSettings,
    CppCompilationDatabase,
    CppCompilationDatabaseFileHashes,
    CppCompilationDatabaseFileTargets,
    CppCompilationDatabasesMap,
    CppIdeFeaturesState,
    CppIdeFeaturesTarget,
    find_cipd_installed_exe_path,
)

from pw_ide.exceptions import (
    BadCompDbException,
    InvalidTargetException,
    MissingCompDbException,
)

from pw_ide.python import PythonPaths

from pw_ide.settings import (
    PigweedIdeSettings,
    SupportedEditor,
)

from pw_ide.vscode import (
    build_extension as build_vscode_extension,
    VscSettingsManager,
    VscSettingsType,
)

_LOG = logging.getLogger(__package__)
env = pigweed_environment()


def _inject_reporter(func):
    """Inject a status reporter instance based on selected output type."""

    def wrapped(*args, **kwargs):
        output = kwargs.pop('output', 'stdout')
        reporter = StatusReporter()

        if output == 'log':
            reporter = LoggingStatusReporter(_LOG)

        kwargs['reporter'] = reporter
        return func(*args, **kwargs)

    # Hoist the decorated function's docstring onto the new function so that
    # we can still access it to auto-generate CLI documentation.
    wrapped.__doc__ = func.__doc__
    return wrapped


def _make_working_dir(
    reporter: StatusReporter, settings: PigweedIdeSettings, quiet: bool = False
) -> None:
    if not settings.working_dir.exists():
        settings.working_dir.mkdir(parents=True)
        if not quiet:
            reporter.new(
                'Initialized the Pigweed IDE working directory at '
                f'{settings.working_dir}'
            )


def _report_unrecognized_editor(reporter: StatusReporter, editor: str) -> None:
    supported_editors = ', '.join(sorted([ed.value for ed in SupportedEditor]))
    reporter.wrn(f'Unrecognized editor: {editor}')
    reporter.wrn('This may not be an automatically-supported editor.')
    reporter.wrn(f'Automatically-supported editors: {supported_editors}')


@_inject_reporter
def cmd_sync(
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Setup or sync your Pigweed project IDE features.

    This will automatically set up your development environment with all the
    features that Pigweed IDE supports, with sensible defaults.

    At minimum, this command will create the .pw_ide working directory and
    create settings files for all supported editors. Projects can define
    additional setup steps in .pw_ide.yaml.

    When new IDE features are introduced in the future (either by Pigweed or
    your downstream project), you can re-run this command to set up the new
    features. It will not overwrite or break any of your existing configuration.

    All commands will be run from the `PW_PROJECT_ROOT` directory
    """
    reporter.info('Syncing pw_ide...')
    _make_working_dir(reporter, pw_ide_settings)

    for command in pw_ide_settings.sync:
        _LOG.debug("Running: %s", command)
        subprocess.run(
            shlex.split(command), cwd=pigweed_root(use_env_vars=False)[0]
        )

    if pw_ide_settings.editor_enabled('vscode'):
        cmd_vscode()

    reporter.info('Done')


@_inject_reporter
def cmd_setup(
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Deprecated! Please use `pw ide sync`."""
    reporter.wrn(
        "The `setup` command is now `sync`. Next time, run `pw ide sync`."
    )
    cmd_sync(reporter, pw_ide_settings)


@_inject_reporter
def cmd_vscode(
    include: list[VscSettingsType] | None = None,
    exclude: list[VscSettingsType] | None = None,
    build_extension: bool = False,
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Configure support for Visual Studio Code.

    This will replace your current Visual Studio Code (VSC) settings for this
    project (in ``.vscode/settings.json``, etc.) with the following sets of
    settings, in order:

    - The Pigweed default settings
    - Your project's settings, if any (in ``.vscode/pw_project_settings.json``)
    - Your personal settings, if any (in ``.vscode/pw_user_settings.json``)

    In other words, settings files lower on the list can override settings
    defined by those higher on the list. Settings defined in the sources above
    are not active in VSC until they are merged and output to the current
    settings file by running:

    .. code-block:: bash

       pw ide vscode

    Refer to the Visual Studio Code documentation for more information about
    these settings: https://code.visualstudio.com/docs/getstarted/settings

    This command also manages VSC tasks (``.vscode/tasks.json``) and extensions
    (``.vscode/extensions.json``). You can explicitly control which of these
    settings types ("settings", "tasks", and "extensions") is modified by
    this command by using the ``--include`` or ``--exclude`` options.

    Your current VSC settings will never change unless you run ``pw ide``
    commands. Since the current VSC settings are an artifact built from the
    three settings files described above, you should avoid manually editing
    that file; it will be replaced the next time you run ``pw ide vscode``. A
    backup of your previous settings file will be made, and you can diff it
    against the new file to see what changed.

    These commands will never modify your VSC user settings, which are
    stored outside of the project repository and apply globally to all VSC
    instances.

    The settings files are system-specific and shouldn't be checked into the
    repository, except for the project settings (those with ``pw_project_``),
    which can be used to define consistent settings for everyone working on the
    project.

    Note that support for VSC can be disabled at the project level or the user
    level by adding the following to .pw_ide.yaml or .pw_ide.user.yaml
    respectively:

    .. code-block:: yaml

        editors:
          vscode: false

    Likewise, it can be enabled by setting that value to true. It is enabled by
    default.
    """
    if build_extension:
        reporter.info('Building the Visual Studio Code extension')

        try:
            build_vscode_extension(Path(env.PW_ROOT))
        except subprocess.CalledProcessError:
            reporter.err("Failed! See output for more info.")
        else:
            reporter.ok('Built successfully!')

    if not pw_ide_settings.editor_enabled('vscode'):
        reporter.wrn(
            'Visual Studio Code support is disabled in settings! If this is '
            'unexpected, see this page for information on enabling support: '
            'https://pigweed.dev/pw_ide/'
            '#pw_ide.settings.PigweedIdeSettings.editors'
        )
        sys.exit(1)

    vsc_manager = VscSettingsManager(pw_ide_settings)

    if include is None and exclude is None:
        include_set = set(VscSettingsType.all())
        exclude_set: Set[VscSettingsType] = set()

    elif include is None:
        include_set = set(VscSettingsType.all())
        exclude_set = set(exclude if exclude is not None else [])

    elif exclude is None:
        include_set = set(include if include is not None else [])
        exclude_set = set()

    else:
        include_set = set(include if include is not None else [])
        exclude_set = set(exclude if exclude is not None else [])

    types_to_update = cast(
        list[VscSettingsType], tuple(include_set - exclude_set)
    )

    for settings_type in types_to_update:
        prev_settings_str = ''
        prev_settings_hash = ''
        active_settings_existed = vsc_manager.active(settings_type).is_present()

        if active_settings_existed:
            prev_settings = vsc_manager.active(settings_type)
            prev_settings_str = str(prev_settings)
            prev_settings_hash = prev_settings.hash()

        with vsc_manager.active(settings_type).build() as active_settings:
            vsc_manager.default(settings_type).sync_to(active_settings)
            vsc_manager.project(settings_type).sync_to(active_settings)
            vsc_manager.user(settings_type).sync_to(active_settings)

        new_settings = vsc_manager.active(settings_type)
        new_settings_str = str(new_settings)
        new_settings_hash = new_settings.hash()
        settings_changed = new_settings_hash != prev_settings_hash

        _LOG.debug(
            'VS Code %s prev hash: %s',
            settings_type.name.lower(),
            prev_settings_hash,
        )
        _LOG.debug(
            'VS Code %s curr hash: %s',
            settings_type.name.lower(),
            new_settings_hash,
        )

        if settings_changed:
            verb = 'Updated' if active_settings_existed else 'Created'
            reporter.new(
                f'{verb} Visual Studio Code active ' f'{settings_type.value}'
            )

            print_diff(
                prev_settings_str.splitlines(True),
                new_settings_str.splitlines(True),
                fromfile=f"{settings_type.value}.json",
                tofile=f"{settings_type.value}.json",
                indent=8,
            )


# Elements:
# - The root search path of the compilation database
#     (essentially, the root directory we started searching from to find it)
# - The compilation database file path
# - The target inference pattern to use
# TODO: https://pwbug.dev/349522832 - This should be a TypedDict or class, but
# doing so may require some careful refactoring.
CompDbPathsData = tuple[Path, Path, str]


def _process_compdbs_from_settings(
    pw_ide_settings: PigweedIdeSettings,
) -> list[CompDbPathsData]:
    """Process comp DBs found in the search path(s) defined in settings."""

    # Find all compilation databases in the paths defined in settings, and
    # associate them with their target inference pattern.
    compdb_paths_data: list[CompDbPathsData] = []
    settings_search_paths = pw_ide_settings.compdb_search_paths

    # Get all compdb_search_paths entries from settings
    for search_path_glob, target_inference in settings_search_paths:
        # Expand the search path globs to get all concrete search paths
        for search_path in (Path(p) for p in iglob(str(search_path_glob))):
            # Search each path for compilation database files
            for compdb_file in search_path.rglob(str(COMPDB_FILE_NAME)):
                compdb_paths_data.append(
                    (Path(search_path), compdb_file, target_inference)
                )

    return compdb_paths_data


def _process_compdbs_from_files(
    pw_ide_settings: PigweedIdeSettings,
    compdb_file_paths: list[Path],
) -> list[CompDbPathsData]:
    """Process comp DBs from provided file path(s)."""

    target_inference = pw_ide_settings.target_inference

    return [
        (file_path.parent.resolve(), file_path, target_inference)
        for file_path in compdb_file_paths
    ]


def _process_compdbs(  # pylint: disable=too-many-locals
    reporter: StatusReporter,
    pw_ide_settings: PigweedIdeSettings,
    compbd_file_paths: list[Path] | None = None,
    always_output_new: bool = False,
):
    """Find and process compilation databases in the project.

    This essentially does four things:
    - Find all the compilation databases it can in the build directory
    - For any databases we've seen before and are unchanged, skip them
    - For any we haven't seed before or are changed, process them
    - Save the state to disk so that other commands can examine/change targets
    """

    state = CppIdeFeaturesState(pw_ide_settings)

    # If a compilation database was seen before and is unchanged, or if it's new
    # and we process it, it will end up in the new hashes dict. If we saw it
    # in the past but it no longer exists, it will not move over to the new
    # hashes dict.
    prev_compdb_hashes = state.compdb_hashes
    new_compdb_hashes: CppCompilationDatabaseFileHashes = {}
    prev_compdb_targets = state.compdb_targets
    new_compdb_targets: CppCompilationDatabaseFileTargets = {}

    targets: list[CppIdeFeaturesTarget] = []
    num_new_unprocessed_targets = 0
    num_new_processed_targets = 0
    num_carried_over_targets = 0
    num_removed_targets = len(state.targets.values())

    unprocessed_compdb_files: list[Path] = []
    processed_compdb_files: list[Path] = []

    # Associate processed compilation databases with their original sources
    all_processed_compdbs: dict[Path, CppCompilationDatabasesMap] = {}

    compdb_paths_data = (
        _process_compdbs_from_settings(pw_ide_settings)
        if compbd_file_paths is None
        else _process_compdbs_from_files(pw_ide_settings, compbd_file_paths)
    )

    for (
        compdb_root_dir,
        compdb_file_path,
        target_inference,
    ) in compdb_paths_data:
        # Load the compilation database
        try:
            compdb = CppCompilationDatabase.load(
                compdb_to_load=compdb_file_path,
                root_dir=compdb_root_dir,
                target_inference=target_inference,
            )
        except MissingCompDbException:
            reporter.err(f'File not found: {str(compdb_file_path)}')
            sys.exit(1)
        # TODO(chadnorvell): Recover more gracefully from errors.
        except BadCompDbException:
            reporter.err(
                'File does not match compilation database format: '
                f'{str(compdb_file_path)}'
            )
            sys.exit(1)

        # Check the hash of the compilation database against our cache of
        # database hashes. Have we see this before and is the hash the same?
        # Then we can skip this database.
        if (
            compdb_file_path in prev_compdb_hashes
            and compdb.file_hash == prev_compdb_hashes[compdb_file_path]
        ):
            # Store this hash in the new hash registry.
            new_compdb_hashes[compdb_file_path] = compdb.file_hash
            # Copy the targets associated with this file...
            new_compdb_targets[compdb_file_path] = prev_compdb_targets[
                compdb_file_path
            ]
            # ... and add them to the targets list.
            targets.extend(new_compdb_targets[compdb_file_path])
            num_carried_over_targets += len(
                new_compdb_targets[compdb_file_path]
            )
            num_removed_targets -= len(new_compdb_targets[compdb_file_path])
            continue

        # We haven't seen this database before. Process it.
        processed_compdbs = compdb.process(
            settings=pw_ide_settings,
            path_globs=pw_ide_settings.clangd_query_drivers(
                find_cipd_installed_exe_path("clang++")
            ),
            always_output_new=always_output_new,
        )

        # The source database doesn't actually need processing, so use it as is.
        if processed_compdbs is None:
            # Infer the name of the target from the path
            name = '_'.join(
                compdb_file_path.relative_to(compdb_root_dir).parent.parts
            )

            target = CppIdeFeaturesTarget(
                name=name,
                compdb_file_path=compdb_file_path,
                num_commands=len(
                    CppCompilationDatabase.load(
                        compdb_file_path, compdb_root_dir
                    )
                ),
            )

            # An unprocessed database will have only one target.
            new_compdb_targets[compdb_file_path] = [target]
            unprocessed_compdb_files.append(compdb_file_path)
            targets.append(target)
            num_new_unprocessed_targets += 1

            # Remember that we've seen this database.
            new_compdb_hashes[compdb_file_path] = compdb.file_hash

        else:
            # We need to use the processed databases, so store them for writing.
            # We'll add the targets associated with the processed databases
            # later.
            all_processed_compdbs[compdb_file_path] = processed_compdbs
            processed_compdb_files.append(compdb_file_path)

    if len(all_processed_compdbs) > 0:
        # Merge into one map of target names to compilation database.
        merged_compdbs = CppCompilationDatabasesMap.merge(
            *all_processed_compdbs.values()
        )

        # Write processed databases to files.
        try:
            merged_compdbs.write()
        except TypeError:
            reporter.err('Could not serialize file to JSON!')
            reporter.wrn('pw_ide state will not be persisted.')
            return False

        # Grab the target and file info from the processed databases.
        for target_name, compdb in merged_compdbs.items():
            target = CppIdeFeaturesTarget(
                name=target_name,
                compdb_file_path=cast(Path, compdb.file_path),
                num_commands=len(compdb),
            )

            targets.append(target)
            num_new_processed_targets += 1

            if (
                source := cast(Path, compdb.source_file_path)
            ) not in new_compdb_targets:
                new_compdb_targets[source] = [target]
                new_compdb_hashes[source] = cast(str, compdb.source_file_hash)
            else:
                new_compdb_targets[source].append(target)

    # Write out state.
    targets_dict = {target_data.name: target_data for target_data in targets}
    state.targets = targets_dict
    state.compdb_hashes = new_compdb_hashes
    state.compdb_targets = new_compdb_targets

    # If the current target is no longer valid, unset it.
    if (
        state.current_target is not None
        and state.current_target.name not in targets_dict
    ):
        state.current_target = None

    num_total_targets = len(targets)
    num_new_targets = num_new_processed_targets + num_new_unprocessed_targets

    # Report the results.
    # Return True if anything meaningful changed as a result of the processing.
    # If the new state is essentially identical to the old state, return False
    # so the caller can avoid needlessly updating anything else.
    if num_new_targets > 0 or num_removed_targets > 0:
        found_compdb_text = (
            f'Found {len(compdb_paths_data)} compilation database'
        )

        if len(compdb_paths_data) > 1:
            found_compdb_text += 's'

        reporter.ok(found_compdb_text)

        reporter_lines = []

        if len(unprocessed_compdb_files) > 0:
            reporter_lines.append(
                f'Linked {len(unprocessed_compdb_files)} '
                'unmodified compilation databases'
            )

        if len(processed_compdb_files) > 0:
            working_dir_path = pw_ide_settings.working_dir.relative_to(
                Path(env.PW_PROJECT_ROOT)
            )
            reporter_lines.append(
                f'Processed {len(processed_compdb_files)} to working dir at '
                f'{working_dir_path}'
            )

        if len(reporter_lines) > 0:
            reporter_lines.extend(
                [
                    f'{num_total_targets} targets are now available '
                    f'({num_new_targets} are new, '
                    f'{num_removed_targets} were removed)',
                ]
            )

            reporter.new(reporter_lines)

        return True
    return False


class TryAgainException(Exception):
    """A signal to retry an action."""


@_inject_reporter
def cmd_cpp(  # pylint: disable=too-many-arguments, too-many-locals, too-many-branches, too-many-statements
    should_list_targets: bool,
    should_get_target: bool,
    target_to_set: str | None,
    process: bool = False,
    process_files: list[Path] | None = None,
    use_default_target: bool = False,
    clangd_command: bool = False,
    clangd_command_system: str | None = None,
    should_try_compdb_gen_cmd: bool = True,
    reporter: StatusReporter = StatusReporter(),
    pw_ide_settings: PigweedIdeSettings = PigweedIdeSettings(),
) -> None:
    """Configure C/C++ code intelligence support.

    Code intelligence can be provided by clangd or other language servers that
    use the clangd compilation database format, defined at:
    https://clang.llvm.org/docs/JSONCompilationDatabase.html

    Pigweed projects define their build configuration(s) via a build system,
    usually GN, Bazel, or CMake. Based on build configurations, the build
    system generates commands to compile each translation unit in the project.
    clangd uses those commands to parse the build graph and provide rich code
    intelligence.

    Pigweed projects often target multiple devices & architectures, and use
    multiple compiler toolchains. As a result, there may be more than one way
    to compile each translation unit. Your build system ensures that it only
    invokes a single compiler command for each translation unit which is
    consistent with the toolchain and target appropriate to that build, which
    we refer to as a "target toolchain".

    We need to do the same thing with the compilation database that clangd uses.
    We handle this by:

    - Processing the compilation database produced the build system into
      multiple internally-consistent compilation databases, one for each
      target toolchain.

    - Providing commands to select which target toolchain you want to use for
      code analysis.

    Refer to the Pigweed documentation or your build system's documentation to
    learn how to produce a clangd compilation database. Once you have one, run
    this command to process it:

    .. code-block:: bash

        pw ide cpp --process-files {path to compile_commands.json}

    When using GN and CMake, ``pw_ide`` can search your build directory (or
    other directories specified in settings) for compilation databases with:

    .. code-block:: bash

       pw ide cpp --process

    You can now examine the target toolchains that are available to you:

    .. code-block:: bash

        pw ide cpp --list

    ... and select the target toolchain you want to use:

    .. code-block:: bash

        pw ide cpp --set host_clang

    As long as your editor or language server plugin is properly configured, you
    will now get code intelligence features relevant to that particular target
    toolchain.

    You can see what target toolchain is selected by running:

    .. code-block:: bash

        pw ide cpp

    Whenever you switch to a target toolchain you haven't used before, clangd
    will index the build, which may take several minutes. This process is not
    blocking, so you can take advantage of code analysis immediately even while
    the indexing is in progress. These indexes are cached, so you can switch
    between targets without re-indexing each time.

    If your build configuration changes significantly (e.g. you add a new file
    to the project), you will need to re-process the compilation database for
    that change to be recognized and manifested in the target toolchain. Your
    target toolchain selection will not change, and your index will only need to
    be incrementally updated.

    You can generate the clangd command your editor needs to run with:

    .. code-block:: bash

        pw ide cpp --clangd-command

    If your editor uses JSON for configuration, you can export the same command
    in that format:

    .. code-block:: bash

        pw ide cpp --clangd-command-for json
    """
    _make_working_dir(reporter, pw_ide_settings, quiet=True)

    # If true, no arguments were provided so we do the default behavior.
    default = True

    state = CppIdeFeaturesState(pw_ide_settings)

    if process or process_files is not None:
        default = False
        _process_compdbs(
            reporter, pw_ide_settings, compbd_file_paths=process_files
        )

        if state.current_target is None:
            use_default_target = True

    if use_default_target:
        defined_default = pw_ide_settings.default_target

        if defined_default is None and state.max_commands_target is None:
            reporter.err(
                'Can\'t use default target toolchain because none is defined!'
            )
            reporter.wrn('Have you processed a compilation database yet?')
            sys.exit(1)
        else:
            max_commands_target = cast(
                CppIdeFeaturesTarget, state.max_commands_target
            )

        default_target = (
            defined_default
            if defined_default is not None
            else max_commands_target.name
        )

        if state.current_target != default:
            target_to_set = default_target

    if target_to_set is not None:
        default = False
        reporter.info(
            f'Setting C/C++ analysis target toolchain to: {target_to_set}'
        )

        try:
            CppIdeFeaturesState(
                pw_ide_settings
            ).current_target = state.targets.get(target_to_set, None)

            if str(CppIdeFeaturesState(pw_ide_settings).current_target) != str(
                target_to_set
            ):
                reporter.err(
                    f'Failed to set target toolchain to {target_to_set}!'
                )
                reporter.wrn(
                    [
                        'You have tried to set a target toolchain '
                        'that is not available.',
                        'Run `pw ide cpp --list` to show available '
                        'target toolchains.',
                        f'If you expected {target_to_set} to be in that list',
                        'and it is not, you may need to use your build system',
                        'generate a compilation database.',
                    ]
                )

                if (
                    should_try_compdb_gen_cmd
                    and pw_ide_settings.compdb_gen_cmd is not None
                ):
                    raise TryAgainException

                sys.exit(1)

        except TryAgainException:
            if pw_ide_settings.compdb_gen_cmd is not None:
                reporter.info(
                    'Will try to generate a compilation database with: '
                    f'{pw_ide_settings.compdb_gen_cmd}'
                )

                subprocess.run(shlex.split(pw_ide_settings.compdb_gen_cmd))

                cmd_cpp(
                    should_list_targets=should_list_targets,
                    should_get_target=should_get_target,
                    target_to_set=target_to_set,
                    process=process,
                    use_default_target=use_default_target,
                    clangd_command=clangd_command,
                    clangd_command_system=clangd_command_system,
                    should_try_compdb_gen_cmd=False,
                )
        except InvalidTargetException:
            reporter.err(
                f'Invalid target toolchain! {target_to_set} not among the '
                'defined target toolchains.'
            )
            sys.exit(1)
        except MissingCompDbException:
            reporter.err(
                f'File not found for target toolchain! {target_to_set}'
            )
            sys.exit(1)

        reporter.new(
            'Set C/C++ language server analysis target toolchain to: '
            f'{CppIdeFeaturesState(pw_ide_settings).current_target}'
        )

    if clangd_command:
        default = False
        reporter.info(
            [
                'Command to run clangd with Pigweed paths:',
                ClangdSettings(pw_ide_settings).command(),
            ]
        )

    if clangd_command_system is not None:
        default = False
        reporter.info(
            [
                'Command to run clangd with Pigweed paths for '
                f'{clangd_command_system}:',
                ClangdSettings(pw_ide_settings).command(clangd_command_system),
            ]
        )

    if should_list_targets:
        default = False
        targets_list_status = [
            'C/C++ target toolchains available for language server analysis:'
        ]

        for target in sorted(CppIdeFeaturesState(pw_ide_settings).targets):
            targets_list_status.append(f'\t{target}')

        reporter.info(targets_list_status)

    if should_get_target or default:
        current_target = CppIdeFeaturesState(pw_ide_settings).current_target
        name = 'None' if current_target is None else current_target.name

        reporter.info(
            'Current C/C++ language server analysis '
            f'target toolchain: {name}'
        )


def install_py_module_as_editable(
    module_name: str,
    reporter: StatusReporter,
) -> None:
    """Install a Pigweed Python module in editable mode."""
    reporter.info(f'Installing {module_name} as an editable module')
    try:
        site_packages_path = [
            path for path in sys.path if 'site-packages' in path
        ][0]
    except IndexError:
        reporter.err(f'Could not find {module_name} in the Python path!')
        sys.exit(1)

    reporter.info(f'Found {module_name} at: {site_packages_path}')
    shutil.rmtree(Path(site_packages_path) / module_name)
    src_path = Path(env.PW_ROOT, module_name, 'py')

    try:
        subprocess.run(
            [
                'pip',
                'install',
                '--no-deps',
                '-e',
                str(src_path),
            ],
            check=True,
            stdout=subprocess.PIPE,
        )
    except subprocess.CalledProcessError:
        reporter.err(
            [
                f'Failed to install {module_name}!',
                'You may need to re-bootstrap',
            ]
        )

        sys.exit(1)

    reporter.new('Success!')
    reporter.wrn('Note that running bootstrap or building will reverse this.')


@_inject_reporter
def cmd_python(
    should_print_venv: bool,
    install_editable: str | None = None,
    reporter: StatusReporter = StatusReporter(),
) -> None:
    """Configure Python code intelligence support.

    You can generate the path to the Python virtual environment interpreter that
    your editor/language server should use with:

    .. code-block:: bash

       pw ide python --venv

    When working on Pigweed's Python modules, it can be convenient to install
    them in editable mode to instantly realize code changes. You can do this by
    running:

    .. code-block:: bash

       pw ide python --install-editable pw_{module name}

    Just note that running bootstrap or building will override this.
    """
    # If true, no arguments were provided and we should do the default
    # behavior.
    default = True

    if install_editable is not None:
        default = False
        install_py_module_as_editable(install_editable, reporter)

    if should_print_venv or default:
        reporter.info(
            [
                'Location of the Pigweed Python virtual environment:',
                str(PythonPaths().interpreter),
            ]
        )



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import sys
import b
import c
import os
import a

class Tree:
  branches: int
  height_cubits: float

def do_something_really_interesting_with_a_lot_of_arguments(arg_foo,arg_a:str, arg_b:int =0, **kwargs)->None:
       call_some_func()
       print(
         "a str"
       )



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Mouse handler fuctions."""

from prompt_toolkit.application.current import get_app
from prompt_toolkit.filters import has_focus
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType


def focus_handler(container, mouse_event: MouseEvent):
    """Focus container on click."""
    if not has_focus(container)():
        if mouse_event.event_type == MouseEventType.MOUSE_UP:
            get_app().layout.focus(container)
            return None
    return NotImplemented



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""C++-related checks."""

import logging
from pathlib import Path
import re
from typing import Callable, Iterable, Iterator

from pw_presubmit.presubmit import (
    Check,
    filter_paths,
)
from pw_presubmit.presubmit_context import PresubmitContext
from pw_presubmit import (
    build,
    format_code,
    presubmit_context,
)

_LOG: logging.Logger = logging.getLogger(__name__)


def _fail(ctx, error, path):
    ctx.fail(error, path=path)
    with open(ctx.failure_summary_log, 'a') as outs:
        print(f'{path}\n{error}\n', file=outs)


@filter_paths(endswith=format_code.CPP_HEADER_EXTS, exclude=(r'\.pb\.h$',))
def pragma_once(ctx: PresubmitContext) -> None:
    """Presubmit check that ensures all header files contain '#pragma once'."""

    ctx.paths = presubmit_context.apply_exclusions(ctx)

    for path in ctx.paths:
        _LOG.debug('Checking %s', path)
        with path.open() as file:
            for line in file:
                if line.startswith('#pragma once'):
                    break
            else:
                _fail(ctx, '#pragma once is missing!', path=path)


def include_guard_check(
    guard: Callable[[Path], str] | None = None,
    allow_pragma_once: bool = True,
) -> Check:
    """Create an include guard check.

    Args:
        guard: Callable that generates an expected include guard name for the
            given Path. If None, any include guard is acceptable, as long as
            it's consistent between the '#ifndef' and '#define' lines.
        allow_pragma_once: Whether to allow headers to use '#pragma once'
            instead of '#ifndef'/'#define'.
    """

    def stripped_non_comment_lines(iterable: Iterable[str]) -> Iterator[str]:
        """Yield non-comment non-empty lines from a C++ file."""
        multi_line_comment = False
        for line in iterable:
            line = line.strip()
            if not line:
                continue
            if line.startswith('//'):
                continue
            if line.startswith('/*'):
                multi_line_comment = True
            if multi_line_comment:
                if line.endswith('*/'):
                    multi_line_comment = False
                continue
            yield line

    def check_path(ctx: PresubmitContext, path: Path) -> None:
        """Check if path has a valid include guard."""

        _LOG.debug('checking %s', path)
        expected: str | None = None
        if guard:
            expected = guard(path)
            _LOG.debug('expecting guard %r', expected)

        with path.open() as ins:
            iterable = stripped_non_comment_lines(ins)
            first_line = next(iterable, '')
            _LOG.debug('first line %r', first_line)

            if allow_pragma_once and first_line.startswith('#pragma once'):
                _LOG.debug('found %r', first_line)
                return

            if expected:
                ifndef_expected = f'#ifndef {expected}'
                if not re.match(rf'^#ifndef {expected}$', first_line):
                    _fail(
                        ctx,
                        'Include guard is missing! Expected: '
                        f'{ifndef_expected!r}, Found: {first_line!r}',
                        path=path,
                    )
                    return

            else:
                match = re.match(
                    r'^#\s*ifndef\s+([a-zA-Z_][a-zA-Z_0-9]*)$',
                    first_line,
                )
                if not match:
                    _fail(
                        ctx,
                        'Include guard is missing! Expected "#ifndef" line, '
                        f'Found: {first_line!r}',
                        path=path,
                    )
                    return
                expected = match.group(1)

            second_line = next(iterable, '')
            _LOG.debug('second line %r', second_line)

            if not re.match(rf'^#\s*define\s+{expected}$', second_line):
                define_expected = f'#define {expected}'
                _fail(
                    ctx,
                    'Include guard is missing! Expected: '
                    f'{define_expected!r}, Found: {second_line!r}',
                    path=path,
                )
                return

            _LOG.debug('passed')

    @filter_paths(endswith=format_code.CPP_HEADER_EXTS, exclude=(r'\.pb\.h$',))
    def include_guard(ctx: PresubmitContext):
        """Check that all header files contain an include guard."""
        ctx.paths = presubmit_context.apply_exclusions(ctx)
        for path in ctx.paths:
            check_path(ctx, path)

    return include_guard


@Check
def asan(ctx: PresubmitContext) -> None:
    """Test with the address sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'asan')


@Check
def msan(ctx: PresubmitContext) -> None:
    """Test with the memory sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'msan')


@Check
def tsan(ctx: PresubmitContext) -> None:
    """Test with the thread sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'tsan')


@Check
def ubsan(ctx: PresubmitContext) -> None:
    """Test with the undefined behavior sanitizer."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'ubsan')


@Check
def runtime_sanitizers(ctx: PresubmitContext) -> None:
    """Test with the address, thread, and undefined behavior sanitizers."""
    build.gn_gen(ctx)
    build.ninja(ctx, 'runtime_sanitizers')


def all_sanitizers():
    # TODO: b/234876100 - msan will not work until the C++ standard library
    # included in the sysroot has a variant built with msan.
    return [asan, tsan, ubsan, runtime_sanitizers]



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Writes a formatted BUILD.gn _file."""

import os
import subprocess

from datetime import datetime
from pathlib import PurePath, PurePosixPath
from types import TracebackType
from typing import IO, Iterable, Set, Type

from pw_build.gn_target import GnTarget

COPYRIGHT_HEADER = f'''
# Copyright {datetime.now().year} The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

# DO NOT MANUALLY EDIT!'''


class MalformedGnError(Exception):
    """Raised when creating a GN object fails."""


class GnWriter:
    """Represents a partial BUILD.gn file being constructed.

    Except for testing , callers should prefer using `GnFile`. That
    object wraps this one, and ensures that the GN file produced includes the
    relevant copyright header and is formatted correctly.
    """

    def __init__(self, file: IO) -> None:
        self._file: IO = file
        self._scopes: list[str] = []
        self._margin: str = ''

    def write_comment(self, comment: str | None = None) -> None:
        """Adds a GN comment.

        Args:
            comment: The comment string to write.
        """
        if not comment:
            self.write('#')
            return
        while len(comment) > 78:
            index = comment.rfind(' ', 0, 78)
            if index < 0:
                break
            self.write(f'# {comment[:index]}')
            comment = comment[index + 1 :]
        self.write(f'# {comment}')

    def write_import(self, gni: str | PurePosixPath) -> None:
        """Adds a GN import.

        Args:
            gni: The source-relative path to a GN import file.
        """
        self.write(f'import("{str(gni)}")')

    def write_target(self, gn_target: GnTarget) -> None:
        """Write a GN target.

        Args:
            target: The GN target data to write.
        """
        if gn_target.origin:
            self.write_comment(f'Generated from {gn_target.origin}')
        self.write(f'{gn_target.type()}("{gn_target.name()}") {{')
        self._margin += '  '

        # inclusive-language: disable
        # See https://gn.googlesource.com/gn/+/master/docs/style_guide.md
        # inclusive-language: enable
        ordered_names = [
            'public',
            'sources',
            'inputs',
            'cflags',
            'include_dirs',
            'defines',
            'public_configs',
            'configs',
            'public_deps',
            'deps',
        ]
        for attr in ordered_names:
            self.write_list(attr, gn_target.attrs.get(attr, []))
        self._margin = self._margin[2:]
        self.write('}')

    def write_list(self, var_name: str, items: Iterable[str]) -> None:
        """Adds a named GN list of the given items, if non-empty.

        Args:
            var_name: The name of the GN list variable.
            items: The list items to write as strings.
        """
        items = list(items)
        if not items:
            return
        self.write(f'{var_name} = [')
        self._margin += '  '
        for item in sorted(items):
            self.write(f'"{str(item)}",')
        self._margin = self._margin[2:]
        self.write(']')

    def write_blank(self) -> None:
        """Adds a blank line."""
        print('', file=self._file)

    def write_preformatted(self, preformatted: str) -> None:
        """Adds text with minimal formatting.

        The only formatting applied to the given text is to strip any leading
        whitespace. This allows calls to be more readable by allowing
        preformatted text to start on a new line, e.g.

            _write_preformatted('''
          preformatted line 1
          preformatted line 2
          preformatted line 3''')

        Args:
            preformatted: The text to write.
        """
        print(preformatted.lstrip(), file=self._file)

    def write_file(self, imports: Set[str], gn_targets: list[GnTarget]) -> None:
        """Write a complete BUILD.gn file.

        Args:
            imports: A list of GNI files needed by targets.
            gn_targets: A list of GN targets to add to the file.
        """
        self.write_import('//build_overrides/pigweed.gni')
        if imports:
            self.write_blank()
        for gni in sorted(list(imports)):
            self.write_import(gni)
        for target in sorted(gn_targets, key=lambda t: t.name()):
            self.write_blank()
            self.write_target(target)

    def write(self, text: str) -> None:
        """Writes to the file, appropriately indented.

        Args:
            text: The text to indent and write.
        """
        print(f'{self._margin}{text}', file=self._file)


class GnFile:
    """Represents an open BUILD.gn file that is formatted on close.

    Typical usage:

        with GnFile('/path/to/BUILD.gn', 'my-package') as build_gn:
          build_gn.write_...

    where "write_..." refers to any of the "write" methods of `GnWriter`.
    """

    def __init__(self, pathname: PurePath) -> None:
        if pathname.name != 'BUILD.gn' and pathname.suffix != '.gni':
            raise MalformedGnError(f'invalid GN filename: {pathname}')
        os.makedirs(pathname.parent, exist_ok=True)
        self._pathname: PurePath = pathname
        self._file: IO
        self._writer: GnWriter

    def __enter__(self) -> GnWriter:
        """Opens the GN file."""
        self._file = open(self._pathname, 'w+')
        self._writer = GnWriter(self._file)
        self._writer.write_preformatted(COPYRIGHT_HEADER)
        file = PurePath(*PurePath(__file__).parts[-2:])
        self._writer.write_comment(
            f'This file was automatically generated by {file}'
        )
        self._writer.write_blank()
        return self._writer

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Closes the GN file and formats it."""
        self._file.close()
        subprocess.check_call(['gn', 'format', self._pathname])



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Creates a new Pigweed module."""

from __future__ import annotations

import abc
import argparse
import dataclasses
from dataclasses import dataclass
import datetime
import difflib
from enum import Enum
import functools
import json
import logging
from pathlib import Path
import re
import sys
from typing import Any, Collection, Iterable, Type

from prompt_toolkit import prompt

from pw_build import generate_modules_lists
import pw_cli.color
import pw_cli.env
from pw_cli.diff import colorize_diff
from pw_cli.status_reporter import StatusReporter

from pw_module.templates import get_template

_COLOR = pw_cli.color.colors()
_LOG = logging.getLogger(__name__)
_PW_ENV = pw_cli.env.pigweed_environment()
_PW_ROOT = _PW_ENV.PW_ROOT

_PIGWEED_LICENSE = f"""
# Copyright {datetime.datetime.now().year} The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.""".lstrip()

_PIGWEED_LICENSE_CC = _PIGWEED_LICENSE.replace('#', '//')

_CREATE = _COLOR.green('create    ')
_REPLACE = _COLOR.green('replace   ')
_UPDATE = _COLOR.yellow('update    ')
_UNCHANGED = _COLOR.blue('unchanged ')
_IDENTICAL = _COLOR.blue('identical ')
_REPORT = StatusReporter()


def _report_write_file(file_path: Path) -> None:
    """Print a notification when a file is newly created or replaced."""
    relative_file_path = str(file_path.relative_to(_PW_ROOT))
    if file_path.is_file():
        _REPORT.info(_REPLACE + relative_file_path)
        return
    _REPORT.new(_CREATE + relative_file_path)


def _report_unchanged_file(file_path: Path) -> None:
    """Print a notification a file was not updated/changed."""
    relative_file_path = str(file_path.relative_to(_PW_ROOT))
    _REPORT.ok(_UNCHANGED + relative_file_path)


def _report_identical_file(file_path: Path) -> None:
    """Print a notification a file is identical."""
    relative_file_path = str(file_path.relative_to(_PW_ROOT))
    _REPORT.ok(_IDENTICAL + relative_file_path)


def _report_edited_file(file_path: Path) -> None:
    """Print a notification a file was modified/edited."""
    relative_file_path = str(file_path.relative_to(_PW_ROOT))
    _REPORT.new(_UPDATE + relative_file_path)


class PromptChoice(Enum):
    """Possible prompt responses."""

    YES = 'yes'
    NO = 'no'
    DIFF = 'diff'


def _prompt_user(message: str, allow_diff: bool = False) -> PromptChoice:
    """Prompt the user for to choose between yes, no and optionally diff.

    If the user presses enter with no text the response is assumed to be NO.
    If the user presses ctrl-c call sys.exit(1).

    Args:
      message: The message to display at the start of the prompt.
      allow_diff: If true add a 'd' to the help text in the prompt line.

    Returns:
      A PromptChoice enum value.
    """
    help_text = '[y/N]'
    if allow_diff:
        help_text = '[y/N/d]'

    try:
        decision = prompt(f'{message} {help_text} ')
    except KeyboardInterrupt:
        sys.exit(1)  # Ctrl-C pressed

    if not decision or decision.lower().startswith('n'):
        return PromptChoice.NO
    if decision.lower().startswith('y'):
        return PromptChoice.YES
    if decision.lower().startswith('d'):
        return PromptChoice.DIFF

    return PromptChoice.NO


def _print_diff(file_name: Path | str, in_text: str, out_text: str) -> None:
    result_diff = list(
        difflib.unified_diff(
            in_text.splitlines(True),
            out_text.splitlines(True),
            f'{file_name}  (original)',
            f'{file_name}  (updated)',
        )
    )
    if not result_diff:
        return
    print()
    print(''.join(colorize_diff(result_diff)))


def _prompt_overwrite(file_path: Path, new_contents: str) -> bool:
    """Returns true if a file should be written, prompts the user if needed."""
    # File does not exist
    if not file_path.is_file():
        return True

    # File exists but is identical.
    old_contents = file_path.read_text(encoding='utf-8')
    if new_contents and old_contents == new_contents:
        _report_identical_file(file_path)
        return False

    file_name = file_path.relative_to(_PW_ROOT)
    # File exists and is different.
    _REPORT.wrn(f'{file_name} already exists.')

    while True:
        choice = _prompt_user('Overwrite?', allow_diff=True)
        if choice == PromptChoice.DIFF:
            _print_diff(file_name, old_contents, new_contents)
        else:
            if choice == PromptChoice.YES:
                return True
            break

    # By default do not overwrite.
    _report_unchanged_file(file_path)
    return False


# TODO(frolv): Adapted from pw_protobuf. Consolidate them.
class _OutputFile:
    DEFAULT_INDENT_WIDTH = 2

    def __init__(self, file: Path, indent_width: int = DEFAULT_INDENT_WIDTH):
        self._file = file
        self._content: list[str] = []
        self._indent_width: int = indent_width
        self._indentation = 0

    def line(self, line: str = '') -> None:
        if line:
            self._content.append(' ' * self._indentation)
            self._content.append(line)
        self._content.append('\n')

    def indent(
        self,
        width: int | None = None,
    ) -> _OutputFile._IndentationContext:
        """Increases the indentation level of the output."""
        return self._IndentationContext(
            self, width if width is not None else self._indent_width
        )

    @property
    def path(self) -> Path:
        return self._file

    @property
    def content(self) -> str:
        return ''.join(self._content)

    def write(self, content: str | None = None) -> None:
        """Write file contents. Prompts the user if necessary.

        Args:
          content: If provided will write this text to the file instead of
              calling self.content.
        """
        output_text = self.content
        if content:
            output_text = content

        if not output_text.endswith('\n'):
            output_text += '\n'

        if _prompt_overwrite(self._file, new_contents=output_text):
            _report_write_file(self._file)
            self._file.write_text(output_text)

    def write_template(self, template_name: str, **template_args) -> None:
        template = get_template(template_name)
        rendered_template = template.render(**template_args)
        self.write(content=rendered_template)

    class _IndentationContext:
        """Context that increases the output's indentation when it is active."""

        def __init__(self, output: _OutputFile, width: int):
            self._output = output
            self._width: int = width

        def __enter__(self):
            self._output._indentation += self._width

        def __exit__(self, typ, value, traceback):
            self._output._indentation -= self._width


class _ModuleName:
    _MODULE_NAME_REGEX = re.compile(
        # Match the two letter character module prefix e.g. 'pw':
        r'^(?P<prefix>[a-zA-Z]{2,})'
        # The rest of the module name consisting of multiple groups of a single
        # underscore followed by alphanumeric characters. This prevents multiple
        # underscores from appearing in a row and the name from ending in a an
        # underscore.
        r'(?P<main>'
        r'(_[a-zA-Z0-9]+)+'
        r')$'
    )

    def __init__(self, prefix: str, main: str, path: Path) -> None:
        self._prefix = prefix
        self._main = main.lstrip('_')  # Remove the leading underscore
        self._path = path

    @property
    def path(self) -> str:
        # Check if there are no parent directories for the full path.
        # Note: This relies on Path('pw_module').parents returning Path('.') for
        # paths that have no parent directories:
        # https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.parents
        if self._path == Path('.'):
            return self.full
        return (self._path / self.full).as_posix()

    @property
    def full(self) -> str:
        return f'{self._prefix}_{self._main}'

    @property
    def prefix(self) -> str:
        return self._prefix

    @property
    def main(self) -> str:
        return self._main

    @property
    def default_namespace(self) -> str:
        return f'{self._prefix}::{self._main}'

    def upper_camel_case(self) -> str:
        return ''.join(s.capitalize() for s in self._main.split('_'))

    @property
    def header_line(self) -> str:
        return '=' * len(self.full)

    def __str__(self) -> str:
        return self.full

    def __repr__(self) -> str:
        return self.full

    @classmethod
    def parse(cls, name: str) -> _ModuleName | None:
        module_path = Path(name)
        module_name = module_path.name
        match = _ModuleName._MODULE_NAME_REGEX.fullmatch(module_name)
        if not match:
            return None

        parts = match.groupdict()
        return cls(parts['prefix'], parts['main'], module_path.parents[0])


@dataclass
class _ModuleContext:
    name: _ModuleName
    dir: Path
    root_build_files: list[_BuildFile]
    sub_build_files: list[_BuildFile]
    build_systems: list[str]
    is_upstream: bool

    def build_files(self) -> Iterable[_BuildFile]:
        yield from self.root_build_files
        yield from self.sub_build_files

    def add_docs_file(self, file: Path):
        for build_file in self.root_build_files:
            build_file.add_docs_source(str(file.relative_to(self.dir)))

    def add_cc_target(self, target: _BuildFile.CcTarget) -> None:
        for build_file in self.root_build_files:
            build_file.add_cc_target(target)

    def add_cc_test(self, target: _BuildFile.CcTarget) -> None:
        for build_file in self.root_build_files:
            build_file.add_cc_test(target)


class _BuildFile:
    """Abstract representation of a build file for a module."""

    @dataclass
    class Target:
        name: str

        # TODO(frolv): Shouldn't be a string list as that's build system
        # specific. Figure out a way to resolve dependencies from targets.
        deps: list[str] = dataclasses.field(default_factory=list)

    @dataclass
    class CcTarget(Target):
        sources: list[Path] = dataclasses.field(default_factory=list)
        headers: list[Path] = dataclasses.field(default_factory=list)

        def rebased_sources(self, rebase_path: Path) -> Iterable[str]:
            return (str(src.relative_to(rebase_path)) for src in self.sources)

        def rebased_headers(self, rebase_path: Path) -> Iterable[str]:
            return (str(hdr.relative_to(rebase_path)) for hdr in self.headers)

    def __init__(self, path: Path, ctx: _ModuleContext):
        self._path = path
        self._ctx = ctx

        self._docs_sources: list[str] = []
        self._cc_targets: list[_BuildFile.CcTarget] = []
        self._cc_tests: list[_BuildFile.CcTarget] = []

    @property
    def path(self) -> Path:
        return self._path

    @property
    def dir(self) -> Path:
        return self._path.parent

    def add_docs_source(self, filename: str) -> None:
        self._docs_sources.append(filename)

    def add_cc_target(self, target: CcTarget) -> None:
        self._cc_targets.append(target)

    def add_cc_test(self, target: CcTarget) -> None:
        self._cc_tests.append(target)

    @property
    def get_license(self) -> str:
        if self._ctx.is_upstream:
            return _PIGWEED_LICENSE
        return ''

    @property
    def docs_sources(self) -> list[str]:
        return self._docs_sources

    @property
    def cc_targets(self) -> list[_BuildFile.CcTarget]:
        return self._cc_targets

    @property
    def cc_tests(self) -> list[_BuildFile.CcTarget]:
        return self._cc_tests

    def relative_file(self, file_path: Path | str) -> str:
        if isinstance(file_path, str):
            return file_path
        return str(file_path.relative_to(self._path.parent))

    def write(self) -> None:
        """Writes the contents of the build file to disk."""
        file = _OutputFile(self._path, self._indent_width())

        if self._ctx.is_upstream:
            file.line(_PIGWEED_LICENSE)
            file.line()

        self._write_preamble(file)

        for target in self._cc_targets:
            file.line()
            self._write_cc_target(file, target)

        for target in self._cc_tests:
            file.line()
            self._write_cc_test(file, target)

        if self._docs_sources:
            file.line()
            self._write_docs_target(file, self._docs_sources)

        file.write()

    @abc.abstractmethod
    def _indent_width(self) -> int:
        """Returns the default indent width for the build file's code style."""

    @abc.abstractmethod
    def _write_preamble(self, file: _OutputFile) -> None:
        """Formats"""

    @abc.abstractmethod
    def _write_cc_target(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        """Defines a C++ library target within the build file."""

    @abc.abstractmethod
    def _write_cc_test(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        """Defines a C++ unit test target within the build file."""

    @abc.abstractmethod
    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: list[str],
    ) -> None:
        """Defines a documentation target within the build file."""


# TODO(frolv): The dict here should be dict[str, '_GnVal'] (i.e. _GnScope),
# but mypy does not yet support recursive types:
# https://github.com/python/mypy/issues/731
_GnVal = bool | int | str | list[str] | dict[str, Any]
_GnScope = dict[str, _GnVal]


class _GnBuildFile(_BuildFile):
    _DEFAULT_FILENAME = 'BUILD.gn'
    _INCLUDE_CONFIG_TARGET = 'public_include_path'

    def __init__(
        self,
        directory: Path,
        ctx: _ModuleContext,
        filename: str = _DEFAULT_FILENAME,
    ):
        super().__init__(directory / filename, ctx)

    def _indent_width(self) -> int:
        return 2

    def _write_preamble(self, file: _OutputFile) -> None:
        # Upstream modules always require a tests target, even if it's empty.
        has_tests = len(self._cc_tests) > 0 or self._ctx.is_upstream

        imports = []

        if self._cc_targets:
            imports.append('$dir_pw_build/target_types.gni')

        if has_tests:
            imports.append('$dir_pw_unit_test/test.gni')

        if self._docs_sources:
            imports.append('$dir_pw_docgen/docs.gni')

        file.line('import("//build_overrides/pigweed.gni")\n')
        for imp in sorted(imports):
            file.line(f'import("{imp}")')

        if self._cc_targets:
            file.line()
            _GnBuildFile._target(
                file,
                'config',
                _GnBuildFile._INCLUDE_CONFIG_TARGET,
                {
                    'include_dirs': ['public'],
                    'visibility': [':*'],
                },
            )

        if has_tests:
            file.line()
            _GnBuildFile._target(
                file,
                'pw_test_group',
                'tests',
                {
                    'tests': list(f':{test.name}' for test in self._cc_tests),
                },
            )

    def _write_cc_target(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        """Defines a GN source_set for a C++ target."""

        target_vars: _GnScope = {}

        if target.headers:
            target_vars['public_configs'] = [
                f':{_GnBuildFile._INCLUDE_CONFIG_TARGET}'
            ]
            target_vars['public'] = list(target.rebased_headers(self.dir))

        if target.sources:
            target_vars['sources'] = list(target.rebased_sources(self.dir))

        if target.deps:
            target_vars['deps'] = target.deps

        _GnBuildFile._target(file, 'pw_source_set', target.name, target_vars)

    def _write_cc_test(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        _GnBuildFile._target(
            file,
            'pw_test',
            target.name,
            {
                'sources': list(target.rebased_sources(self.dir)),
                'deps': target.deps,
            },
        )

    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: list[str],
    ) -> None:
        """Defines a pw_doc_group for module documentation."""
        _GnBuildFile._target(
            file,
            'pw_doc_group',
            'docs',
            {
                'sources': docs_sources,
            },
        )

    @staticmethod
    def _target(
        file: _OutputFile,
        target_type: str,
        name: str,
        args: _GnScope,
    ) -> None:
        """Formats a GN target."""

        file.line(f'{target_type}("{name}") {{')

        with file.indent():
            _GnBuildFile._format_gn_scope(file, args)

        file.line('}')

    @staticmethod
    def _format_gn_scope(file: _OutputFile, scope: _GnScope) -> None:
        """Formats all of the variables within a GN scope to a file.

        This function does not write the enclosing braces of the outer scope to
        support use from multiple formatting contexts.
        """
        for key, val in scope.items():
            if isinstance(val, int):
                file.line(f'{key} = {val}')
                continue

            if isinstance(val, str):
                file.line(f'{key} = {_GnBuildFile._gn_string(val)}')
                continue

            if isinstance(val, bool):
                file.line(f'{key} = {str(val).lower()}')
                continue

            if isinstance(val, dict):
                file.line(f'{key} = {{')
                with file.indent():
                    _GnBuildFile._format_gn_scope(file, val)
                file.line('}')
                continue

            # Format a list of strings.
            # TODO(frolv): Lists of other types?
            assert isinstance(val, list)

            if not val:
                file.line(f'{key} = []')
                continue

            if len(val) == 1:
                file.line(f'{key} = [ {_GnBuildFile._gn_string(val[0])} ]')
                continue

            file.line(f'{key} = [')
            with file.indent():
                for string in sorted(val):
                    file.line(f'{_GnBuildFile._gn_string(string)},')
            file.line(']')

    @staticmethod
    def _gn_string(string: str) -> str:
        """Converts a Python string into a string literal within a GN file.

        Accounts for the possibility of variable interpolation within GN,
        removing quotes if unnecessary:

            "string"           ->  "string"
            "string"           ->  "string"
            "$var"             ->  var
            "$var2"            ->  var2
            "$3var"            ->  "$3var"
            "$dir_pw_foo"      ->  dir_pw_foo
            "$dir_pw_foo:bar"  ->  "$dir_pw_foo:bar"
            "$dir_pw_foo/baz"  ->  "$dir_pw_foo/baz"
            "${dir_pw_foo}"    ->  dir_pw_foo

        """

        # Check if the entire string refers to a interpolated variable.
        #
        # Simple case: '$' followed a single word, e.g. "$my_variable".
        # Note that identifiers can't start with a number.
        if re.fullmatch(r'^\$[a-zA-Z_]\w*$', string):
            return string[1:]

        # GN permits wrapping an interpolated variable in braces.
        # Check for strings of the format "${my_variable}".
        if re.fullmatch(r'^\$\{[a-zA-Z_]\w*\}$', string):
            return string[2:-1]

        return f'"{string}"'


class _BazelBuildFile(_BuildFile):
    _DEFAULT_FILENAME = 'BUILD.bazel'

    def __init__(
        self,
        directory: Path,
        ctx: _ModuleContext,
        filename: str = _DEFAULT_FILENAME,
    ):
        super().__init__(directory / filename, ctx)

    def write(self) -> None:
        """Writes the contents of the build file to disk."""
        file = _OutputFile(self._path)
        file.write_template('BUILD.bazel.jinja', build=self, module=self._ctx)

    def _indent_width(self) -> int:
        return 4

    # TODO(tonymd): Remove these functions once all file types are created with
    # templates.
    def _write_preamble(self, file: _OutputFile) -> None:
        pass

    def _write_cc_target(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        pass

    def _write_cc_test(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        pass

    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: list[str],
    ) -> None:
        pass


class _CmakeBuildFile(_BuildFile):
    _DEFAULT_FILENAME = 'CMakeLists.txt'

    def __init__(
        self,
        directory: Path,
        ctx: _ModuleContext,
        filename: str = _DEFAULT_FILENAME,
    ):
        super().__init__(directory / filename, ctx)

    def write(self) -> None:
        """Writes the contents of the build file to disk."""
        file = _OutputFile(self._path)
        file.write_template(
            'CMakeLists.txt.jinja', build=self, module=self._ctx
        )

    def _indent_width(self) -> int:
        return 2

    # TODO(tonymd): Remove these functions once all file types are created with
    # templates.
    def _write_preamble(self, file: _OutputFile) -> None:
        pass

    def _write_cc_target(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        pass

    def _write_cc_test(
        self,
        file: _OutputFile,
        target: _BuildFile.CcTarget,
    ) -> None:
        pass

    def _write_docs_target(
        self,
        file: _OutputFile,
        docs_sources: list[str],
    ) -> None:
        pass


class _LanguageGenerator:
    """Generates files for a programming language in a new Pigweed module."""

    def __init__(self, ctx: _ModuleContext) -> None:
        self._ctx = ctx

    @abc.abstractmethod
    def create_source_files(self) -> None:
        """Creates the boilerplate source files required by the language."""


class _CcLanguageGenerator(_LanguageGenerator):
    """Generates boilerplate source files for a C++ module."""

    def __init__(self, ctx: _ModuleContext) -> None:
        super().__init__(ctx)

        self._public_dir = ctx.dir / 'public'
        self._headers_dir = self._public_dir / ctx.name.full

    def create_source_files(self) -> None:
        self._headers_dir.mkdir(parents=True, exist_ok=True)

        main_header = self._new_header(self._ctx.name.main)
        main_source = self._new_source(self._ctx.name.main)
        test_source = self._new_source(f'{self._ctx.name.main}_test')

        # TODO(frolv): This could be configurable.
        namespace = self._ctx.name.default_namespace

        main_source.line(
            f'#include "{main_header.path.relative_to(self._public_dir)}"\n'
        )
        main_source.line(f'namespace {namespace} {{\n')
        main_source.line('int magic = 42;\n')
        main_source.line(f'}}  // namespace {namespace}')

        main_header.line(f'namespace {namespace} {{\n')
        main_header.line('extern int magic;\n')
        main_header.line(f'}}  // namespace {namespace}')

        test_source.line(
            f'#include "{main_header.path.relative_to(self._public_dir)}"\n'
        )
        test_source.line('#include "pw_unit_test/framework.h"\n')
        test_source.line(f'namespace {namespace} {{')
        test_source.line('namespace {\n')
        test_source.line(
            f'TEST({self._ctx.name.upper_camel_case()}, GeneratesCorrectly) {{'
        )
        with test_source.indent():
            test_source.line('EXPECT_EQ(magic, 42);')
        test_source.line('}\n')
        test_source.line('}  // namespace')
        test_source.line(f'}}  // namespace {namespace}')

        self._ctx.add_cc_target(
            _BuildFile.CcTarget(
                name=self._ctx.name.full,
                sources=[main_source.path],
                headers=[main_header.path],
            )
        )

        self._ctx.add_cc_test(
            _BuildFile.CcTarget(
                name=f'{self._ctx.name.main}_test',
                deps=[f':{self._ctx.name.full}'],
                sources=[test_source.path],
            )
        )

        main_header.write()
        main_source.write()
        test_source.write()

    def _new_source(self, name: str) -> _OutputFile:
        file = _OutputFile(self._ctx.dir / f'{name}.cc')

        if self._ctx.is_upstream:
            file.line(_PIGWEED_LICENSE_CC)
            file.line()

        return file

    def _new_header(self, name: str) -> _OutputFile:
        file = _OutputFile(self._headers_dir / f'{name}.h')

        if self._ctx.is_upstream:
            file.line(_PIGWEED_LICENSE_CC)

        file.line('#pragma once\n')
        return file


_BUILD_FILES: dict[str, Type[_BuildFile]] = {
    'bazel': _BazelBuildFile,
    'cmake': _CmakeBuildFile,
    'gn': _GnBuildFile,
}

_LANGUAGE_GENERATORS: dict[str, Type[_LanguageGenerator]] = {
    'cc': _CcLanguageGenerator,
}


def _check_module_name(
    module: str,
    is_upstream: bool,
) -> _ModuleName | None:
    """Checks whether a module name is valid."""

    name = _ModuleName.parse(module)
    if not name:
        _LOG.error(
            '"%s" does not conform to the Pigweed module name format', module
        )
        return None

    if is_upstream and name.prefix != 'pw':
        _LOG.error('Modules within Pigweed itself must start with "pw_"')
        return None

    return name


def _create_main_docs_file(ctx: _ModuleContext) -> None:
    """Populates the top-level docs.rst file within a new module."""

    template = get_template('docs.rst.jinja')
    rendered_template = template.render(module=ctx)

    docs_file = _OutputFile(ctx.dir / 'docs.rst')
    ctx.add_docs_file(docs_file.path)
    docs_file.write(content=rendered_template)


def _basic_module_setup(
    module_name: _ModuleName,
    module_dir: Path,
    build_systems: Iterable[str],
    is_upstream: bool,
) -> _ModuleContext:
    """Creates the basic layout of a Pigweed module."""
    module_dir.mkdir(parents=True, exist_ok=True)
    public_dir = module_dir / 'public' / module_name.full
    public_dir.mkdir(parents=True, exist_ok=True)

    ctx = _ModuleContext(
        name=module_name,
        dir=module_dir,
        root_build_files=[],
        sub_build_files=[],
        build_systems=list(build_systems),
        is_upstream=is_upstream,
    )

    ctx.root_build_files.extend(
        _BUILD_FILES[build](module_dir, ctx) for build in ctx.build_systems
    )

    _create_main_docs_file(ctx)

    return ctx


def _add_to_module_metadata(
    project_root: Path,
    module_name: _ModuleName,
    languages: Iterable[str] | None = None,
) -> None:
    """Update sphinx module metadata."""
    module_metadata_file = project_root / 'docs/module_metadata.json'
    metadata_dict = json.loads(module_metadata_file.read_text())

    language_tags = []
    if languages:
        for lang in languages:
            if lang == 'cc':
                language_tags.append('C++')

    # Add the new entry if it doesn't exist
    if module_name.full not in metadata_dict:
        metadata_dict[module_name.full] = dict(
            status='experimental',
            languages=language_tags,
        )

    # Sort by module name.
    sorted_metadata = dict(
        sorted(metadata_dict.items(), key=lambda item: item[0])
    )
    output_text = json.dumps(sorted_metadata, sort_keys=False, indent=2)
    output_text += '\n'

    # Write the file.
    if _prompt_overwrite(module_metadata_file, new_contents=output_text):
        _report_write_file(module_metadata_file)
        module_metadata_file.write_text(output_text)


def _add_to_pigweed_modules_file(
    project_root: Path,
    module_name: _ModuleName,
) -> None:
    modules_file = project_root / 'PIGWEED_MODULES'
    if not modules_file.exists():
        _LOG.error(
            'Could not locate PIGWEED_MODULES file; '
            'your repository may be in a bad state.'
        )
        return

    modules_gni_file = (
        project_root / 'pw_build' / 'generated_pigweed_modules_lists.gni'
    )

    # Cut off the extra newline at the end of the file.
    modules_list = modules_file.read_text().splitlines()
    if module_name.path in modules_list:
        _report_unchanged_file(modules_file)
        return
    modules_list.append(module_name.path)
    modules_list.sort()
    modules_list.append('')
    modules_file.write_text('\n'.join(modules_list))
    _report_edited_file(modules_file)

    generate_modules_lists.main(
        root=project_root,
        modules_list=modules_file,
        modules_gni_file=modules_gni_file,
        mode=generate_modules_lists.Mode.UPDATE,
    )
    _report_edited_file(modules_gni_file)


def _add_to_root_cmakelists(
    project_root: Path,
    module_name: _ModuleName,
) -> None:
    new_line = f'add_subdirectory({module_name.path} EXCLUDE_FROM_ALL)\n'

    path = project_root / 'CMakeLists.txt'
    if not path.exists():
        _LOG.error('Could not locate root CMakeLists.txt file.')
        return

    lines = path.read_text().splitlines(keepends=True)
    if new_line in lines:
        _report_unchanged_file(path)
        return

    add_subdir_start = 0
    while add_subdir_start < len(lines):
        if lines[add_subdir_start].startswith('add_subdirectory'):
            break
        add_subdir_start += 1

    insert_point = add_subdir_start
    while (
        lines[insert_point].startswith('add_subdirectory')
        and lines[insert_point] < new_line
    ):
        insert_point += 1

    lines.insert(insert_point, new_line)
    path.write_text(''.join(lines))
    _report_edited_file(path)


def _project_root() -> Path:
    """Returns the path to the root directory of the current project."""
    project_root = _PW_ENV.PW_PROJECT_ROOT
    if not project_root.is_dir():
        _LOG.error(
            'Expected env var $PW_PROJECT_ROOT to point to a directory, but '
            'found `%s` which is not a directory.',
            project_root,
        )
        sys.exit(1)
    return project_root


def _is_upstream() -> bool:
    """Returns whether this command is being run within Pigweed itself."""
    return _PW_ROOT == _project_root()


_COMMENTS = re.compile(r'\w*#.*$')


def _read_root_owners(project_root: Path) -> Iterable[str]:
    for line in (project_root / 'OWNERS').read_text().splitlines():
        line = _COMMENTS.sub('', line).strip()
        if line:
            yield line


def _create_module(
    module: str,
    languages: Iterable[str],
    build_systems: Iterable[str],
    owners: Collection[str] | None = None,
) -> None:
    project_root = _project_root()
    is_upstream = _is_upstream()

    module_name = _check_module_name(module, is_upstream)
    if not module_name:
        sys.exit(1)

    if not is_upstream:
        _LOG.error(
            '`pw module create` is experimental and does '
            'not yet support downstream projects.'
        )
        sys.exit(1)

    module_dir = project_root / module

    if module_dir.is_dir():
        _REPORT.wrn(f'Directory {module} already exists.')
        if _prompt_user('Continue?') == PromptChoice.NO:
            sys.exit(1)

    if module_dir.is_file():
        _LOG.error(
            'Cannot create module %s as a file of that name already exists',
            module,
        )
        sys.exit(1)

    if owners is not None:
        if len(owners) < 2:
            _LOG.error(
                'New modules must have at least two owners, but only `%s` was '
                'provided.',
                owners,
            )
            sys.exit(1)
        for owner in owners:
            if '@' not in owner:
                _LOG.error(
                    'Owners should be email addresses, but found `%s`', owner
                )
                sys.exit(1)
        root_owners = list(_read_root_owners(project_root))
        if not any(owner in root_owners for owner in owners):
            root_owners_str = '\n'.join(root_owners)
            _LOG.error(
                'Module owners must include at least one root owner, but only '
                '`%s` was provided. Root owners include:\n%s',
                owners,
                root_owners_str,
            )
            sys.exit(1)

    ctx = _basic_module_setup(
        module_name, module_dir, build_systems, is_upstream
    )

    if owners is not None:
        owners_file = module_dir / 'OWNERS'
        owners_text = '\n'.join(sorted(owners))
        owners_text += '\n'
        if _prompt_overwrite(owners_file, new_contents=owners_text):
            _report_write_file(owners_file)
            owners_file.write_text(owners_text)

    try:
        generators = list(_LANGUAGE_GENERATORS[lang](ctx) for lang in languages)
    except KeyError as key:
        _LOG.error('Unsupported language: %s', key)
        sys.exit(1)

    for generator in generators:
        generator.create_source_files()

    for build_file in ctx.build_files():
        build_file.write()

    if is_upstream:
        _add_to_pigweed_modules_file(project_root, module_name)
        _add_to_module_metadata(project_root, module_name, languages)
        if 'cmake' in build_systems:
            _add_to_root_cmakelists(project_root, module_name)

    print()
    _REPORT.new(f'{module_name} created at: {module_dir.relative_to(_PW_ROOT)}')


def register_subcommand(parser: argparse.ArgumentParser) -> None:
    """Registers the module `create` subcommand with `parser`."""
    csv = lambda s: s.split(',')

    def csv_with_choices(choices: list[str], string) -> list[str]:
        chosen_items = list(string.split(','))
        invalid_items = set(chosen_items) - set(choices)
        if invalid_items:
            raise argparse.ArgumentTypeError(
                '\n'
                f'  invalid items: [ {", ".join(invalid_items)} ].\n'
                f'  choose from: [ {", ".join(choices)} ]'
            )

        return chosen_items

    parser.add_argument(
        '--build-systems',
        help=(
            'Comma-separated list of build systems the module supports. '
            f'Options: {", ".join(_BUILD_FILES.keys())}'
        ),
        default=_BUILD_FILES.keys(),
        type=functools.partial(csv_with_choices, _BUILD_FILES.keys()),
    )
    parser.add_argument(
        '--languages',
        help=(
            'Comma-separated list of languages the module will use. '
            f'Options: {", ".join(_LANGUAGE_GENERATORS.keys())}'
        ),
        default=[],
        type=functools.partial(csv_with_choices, _LANGUAGE_GENERATORS.keys()),
    )
    if _is_upstream():
        parser.add_argument(
            '--owners',
            help=(
                'Comma-separated list of emails of the people who will own and '
                'maintain the new module. This list must contain at least two '
                'entries, and at least one user must be a top-level OWNER '
                f'(listed in `{_project_root()}/OWNERS`).'
            ),
            required=True,
            metavar='firstownername@google.com,secondownername@foo.net',
            type=csv,
        )
    parser.add_argument(
        'module', help='Name of the module to create.', metavar='MODULE_NAME'
    )
    parser.set_defaults(func=_create_module)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_chrono_protos/chrono.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dpw_chrono_protos/chrono.proto\x12\tpw.chrono\"q\n\tEpochType\"d\n\x04\x45num\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x13\n\x0fTIME_SINCE_BOOT\x10\x01\x12\x12\n\x0eUTC_WALL_CLOCK\x10\x02\x12\x12\n\x0eGPS_WALL_CLOCK\x10\x03\x12\x12\n\x0eTAI_WALL_CLOCK\x10\x04\"\xa4\x01\n\x0f\x43lockParameters\x12%\n\x1dtick_period_seconds_numerator\x18\x01 \x01(\x05\x12\'\n\x1ftick_period_seconds_denominator\x18\x02 \x01(\x05\x12\x32\n\nepoch_type\x18\x03 \x01(\x0e\x32\x19.pw.chrono.EpochType.EnumH\x00\x88\x01\x01\x42\r\n\x0b_epoch_type\"T\n\tTimePoint\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x34\n\x10\x63lock_parameters\x18\x02 \x01(\x0b\x32\x1a.pw.chrono.ClockParameters\">\n\x12SnapshotTimestamps\x12(\n\ntimestamps\x18\x16 \x03(\x0b\x32\x14.pw.chrono.TimePointB\x08\x42\x06\x43hronob\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_chrono_protos.chrono_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'B\006Chrono'
  _globals['_EPOCHTYPE']._serialized_start=44
  _globals['_EPOCHTYPE']._serialized_end=157
  _globals['_EPOCHTYPE_ENUM']._serialized_start=57
  _globals['_EPOCHTYPE_ENUM']._serialized_end=157
  _globals['_CLOCKPARAMETERS']._serialized_start=160
  _globals['_CLOCKPARAMETERS']._serialized_end=324
  _globals['_TIMEPOINT']._serialized_start=326
  _globals['_TIMEPOINT']._serialized_end=410
  _globals['_SNAPSHOTTIMESTAMPS']._serialized_start=412
  _globals['_SNAPSHOTTIMESTAMPS']._serialized_end=474
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_codegen_protos/codegen_options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0pw_protobuf_codegen_protos/codegen_options.proto\x12\x0bpw.protobuf\"t\n\x0e\x43odegenOptions\x12\x11\n\tmax_count\x18\x01 \x01(\x05\x12\x13\n\x0b\x66ixed_count\x18\x02 \x01(\x08\x12\x10\n\x08max_size\x18\x03 \x01(\x05\x12\x12\n\nfixed_size\x18\x04 \x01(\x08\x12\x14\n\x0cuse_callback\x18\x05 \x01(\x08\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_protobuf_codegen_protos.codegen_options_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_CODEGENOPTIONS']._serialized_start=65
  _globals['_CODEGENOPTIONS']._serialized_end=181
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_log/proto/log.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from pw_protobuf_protos import common_pb2 as pw__protobuf__protos_dot_common__pb2
from pw_tokenizer_proto import options_pb2 as pw__tokenizer__proto_dot_options__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16pw_log/proto/log.proto\x12\x06pw.log\x1a\x1fpw_protobuf_protos/common.proto\x1a pw_tokenizer_proto/options.proto\"\xe4\x01\n\x08LogEntry\x12\x1a\n\x07message\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00H\x01\x88\x01\x01\x12\x12\n\nline_level\x18\x02 \x01(\r\x12\r\n\x05\x66lags\x18\x03 \x01(\r\x12\x13\n\ttimestamp\x18\x04 \x01(\x03H\x00\x12\x1f\n\x15time_since_last_entry\x18\x05 \x01(\x03H\x00\x12\x0f\n\x07\x64ropped\x18\x06 \x01(\r\x12\x14\n\x06module\x18\x07 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x12\n\x04\x66ile\x18\x08 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x14\n\x06thread\x18\t \x01(\x0c\x42\x04\x80\xaf&\x00\x42\x06\n\x04timeB\n\n\x08_message\"\x0c\n\nLogRequest\"P\n\nLogEntries\x12!\n\x07\x65ntries\x18\x01 \x03(\x0b\x32\x10.pw.log.LogEntry\x12\x1f\n\x17\x66irst_entry_sequence_id\x18\x02 \x01(\r\"\xf2\x02\n\nFilterRule\x12=\n\x1blevel_greater_than_or_equal\x18\x01 \x01(\x0e\x32\x18.pw.log.FilterRule.Level\x12\x1b\n\rmodule_equals\x18\x02 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x15\n\rany_flags_set\x18\x03 \x01(\r\x12)\n\x06\x61\x63tion\x18\x04 \x01(\x0e\x32\x19.pw.log.FilterRule.Action\x12\x1b\n\rthread_equals\x18\x05 \x01(\x0c\x42\x04\x80\xaf&\x00\"}\n\x05Level\x12\r\n\tANY_LEVEL\x10\x00\x12\x0f\n\x0b\x44\x45\x42UG_LEVEL\x10\x01\x12\x0e\n\nINFO_LEVEL\x10\x02\x12\x0e\n\nWARN_LEVEL\x10\x03\x12\x0f\n\x0b\x45RROR_LEVEL\x10\x04\x12\x12\n\x0e\x43RITICAL_LEVEL\x10\x05\x12\x0f\n\x0b\x46\x41TAL_LEVEL\x10\x07\"*\n\x06\x41\x63tion\x12\x0c\n\x08INACTIVE\x10\x00\x12\x08\n\x04KEEP\x10\x01\x12\x08\n\x04\x44ROP\x10\x02\"*\n\x06\x46ilter\x12 \n\x04rule\x18\x01 \x03(\x0b\x32\x12.pw.log.FilterRule\"K\n\x10SetFilterRequest\x12\x17\n\tfilter_id\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\x12\x1e\n\x06\x66ilter\x18\x02 \x01(\x0b\x32\x0e.pw.log.Filter\"+\n\x10GetFilterRequest\x12\x17\n\tfilter_id\x18\x01 \x01(\x0c\x42\x04\x80\xaf&\x00\"\x15\n\x13\x46ilterIdListRequest\"/\n\x14\x46ilterIdListResponse\x12\x17\n\tfilter_id\x18\x01 \x03(\x0c\x42\x04\x80\xaf&\x00\x32:\n\x04Logs\x12\x32\n\x06Listen\x12\x12.pw.log.LogRequest\x1a\x12.pw.log.LogEntries0\x01\x32\xc7\x01\n\x07\x46ilters\x12\x39\n\tSetFilter\x12\x18.pw.log.SetFilterRequest\x1a\x12.pw.protobuf.Empty\x12\x35\n\tGetFilter\x12\x18.pw.log.GetFilterRequest\x1a\x0e.pw.log.Filter\x12J\n\rListFilterIds\x12\x1b.pw.log.FilterIdListRequest\x1a\x1c.pw.log.FilterIdListResponseB\x05\x42\x03Logb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_log.proto.log_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'B\003Log'
  _globals['_LOGENTRY'].fields_by_name['message']._options = None
  _globals['_LOGENTRY'].fields_by_name['message']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY'].fields_by_name['module']._options = None
  _globals['_LOGENTRY'].fields_by_name['module']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY'].fields_by_name['file']._options = None
  _globals['_LOGENTRY'].fields_by_name['file']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY'].fields_by_name['thread']._options = None
  _globals['_LOGENTRY'].fields_by_name['thread']._serialized_options = b'\200\257&\000'
  _globals['_FILTERRULE'].fields_by_name['module_equals']._options = None
  _globals['_FILTERRULE'].fields_by_name['module_equals']._serialized_options = b'\200\257&\000'
  _globals['_FILTERRULE'].fields_by_name['thread_equals']._options = None
  _globals['_FILTERRULE'].fields_by_name['thread_equals']._serialized_options = b'\200\257&\000'
  _globals['_SETFILTERREQUEST'].fields_by_name['filter_id']._options = None
  _globals['_SETFILTERREQUEST'].fields_by_name['filter_id']._serialized_options = b'\200\257&\000'
  _globals['_GETFILTERREQUEST'].fields_by_name['filter_id']._options = None
  _globals['_GETFILTERREQUEST'].fields_by_name['filter_id']._serialized_options = b'\200\257&\000'
  _globals['_FILTERIDLISTRESPONSE'].fields_by_name['filter_id']._options = None
  _globals['_FILTERIDLISTRESPONSE'].fields_by_name['filter_id']._serialized_options = b'\200\257&\000'
  _globals['_LOGENTRY']._serialized_start=102
  _globals['_LOGENTRY']._serialized_end=330
  _globals['_LOGREQUEST']._serialized_start=332
  _globals['_LOGREQUEST']._serialized_end=344
  _globals['_LOGENTRIES']._serialized_start=346
  _globals['_LOGENTRIES']._serialized_end=426
  _globals['_FILTERRULE']._serialized_start=429
  _globals['_FILTERRULE']._serialized_end=799
  _globals['_FILTERRULE_LEVEL']._serialized_start=630
  _globals['_FILTERRULE_LEVEL']._serialized_end=755
  _globals['_FILTERRULE_ACTION']._serialized_start=757
  _globals['_FILTERRULE_ACTION']._serialized_end=799
  _globals['_FILTER']._serialized_start=801
  _globals['_FILTER']._serialized_end=843
  _globals['_SETFILTERREQUEST']._serialized_start=845
  _globals['_SETFILTERREQUEST']._serialized_end=920
  _globals['_GETFILTERREQUEST']._serialized_start=922
  _globals['_GETFILTERREQUEST']._serialized_end=965
  _globals['_FILTERIDLISTREQUEST']._serialized_start=967
  _globals['_FILTERIDLISTREQUEST']._serialized_end=988
  _globals['_FILTERIDLISTRESPONSE']._serialized_start=990
  _globals['_FILTERIDLISTRESPONSE']._serialized_end=1037
  _globals['_LOGS']._serialized_start=1039
  _globals['_LOGS']._serialized_end=1097
  _globals['_FILTERS']._serialized_start=1100
  _globals['_FILTERS']._serialized_end=1299
# @@protoc_insertion_point(module_scope)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_build.project_builder_presubmit_runner"""

from __future__ import annotations

import argparse
import fnmatch
import logging
from pathlib import Path


import pw_cli.log
from pw_cli.arguments import (
    print_completions_for_option,
    add_tab_complete_arguments,
)
from pw_presubmit.presubmit import (
    Program,
    Programs,
    Presubmit,
    PresubmitContext,
    Check,
    fetch_file_lists,
)
import pw_presubmit.pigweed_presubmit
from pw_presubmit.build import GnGenNinja, gn_args
from pw_presubmit.presubmit_context import get_check_traces, PresubmitCheckTrace
from pw_presubmit.tools import file_summary

# pw_watch is not required by pw_build, this is an optional feature.
try:
    from pw_watch.argparser import (  # type: ignore
        add_parser_arguments as add_watch_arguments,
    )
    from pw_watch.watch import run_watch, watch_setup  # type: ignore
    from pw_watch.watch_app import WatchAppPrefs  # type: ignore

    PW_WATCH_AVAILABLE = True
except ImportError:
    PW_WATCH_AVAILABLE = False

from pw_build.project_builder import (
    ProjectBuilder,
    run_builds,
    ASCII_CHARSET,
    EMOJI_CHARSET,
)
from pw_build.build_recipe import (
    BuildCommand,
    BuildRecipe,
    UnknownBuildSystem,
    create_build_recipes,
    should_gn_gen,
)
from pw_build.project_builder_argparse import add_project_builder_arguments
from pw_build.project_builder_prefs import ProjectBuilderPrefs


_COLOR = pw_cli.color.colors()
_LOG = logging.getLogger('pw_build')


class PresubmitTraceAnnotationError(Exception):
    """Exception for malformed PresubmitCheckTrace annotations."""


def _pw_package_install_command(package_name: str) -> BuildCommand:
    return BuildCommand(
        command=[
            'pw',
            '--no-banner',
            'package',
            'install',
            package_name,
        ],
    )


def _pw_package_install_to_build_command(
    trace: PresubmitCheckTrace,
) -> BuildCommand:
    """Returns a BuildCommand from a PresubmitCheckTrace."""
    package_name = trace.call_annotation.get('pw_package_install', None)
    if package_name is None:
        raise PresubmitTraceAnnotationError(
            'Missing "pw_package_install" value.'
        )

    return _pw_package_install_command(package_name)


def _bazel_command_args_to_build_commands(
    trace: PresubmitCheckTrace,
) -> list[BuildCommand]:
    """Returns a list of BuildCommands based on a bazel PresubmitCheckTrace."""
    build_steps: list[BuildCommand] = []

    if not 'bazel' in trace.args:
        return build_steps

    bazel_command = list(arg for arg in trace.args if not arg.startswith('--'))
    bazel_options = list(
        arg for arg in trace.args if arg.startswith('--') and arg != '--'
    )
    # Check for `bazel build` or `bazel test`
    if not (
        bazel_command[0].endswith('bazel')
        and bazel_command[1] in ['build', 'test']
    ):
        raise UnknownBuildSystem(
            f'Unable to parse bazel command:\n  {trace.args}'
        )

    bazel_subcommand = bazel_command[1]
    bazel_targets = bazel_command[2:]
    if bazel_subcommand == 'build':
        build_steps.append(
            BuildCommand(
                build_system_command='bazel',
                build_system_extra_args=['build'] + bazel_options,
                targets=bazel_targets,
            )
        )
    if bazel_subcommand == 'test':
        build_steps.append(
            BuildCommand(
                build_system_command='bazel',
                build_system_extra_args=['test'] + bazel_options,
                targets=bazel_targets,
            )
        )
    return build_steps


def _presubmit_trace_to_build_commands(
    ctx: PresubmitContext,
    presubmit_step: Check,
) -> list[BuildCommand]:
    """Convert a presubmit step to a list of BuildCommands.

    Specifically, this handles the following types of PresubmitCheckTraces:

    - pw package installs
    - gn gen followed by ninja
    - bazel commands

    If none of the specific scenarios listed above are found the command args
    are passed along to BuildCommand as is.

    Returns:
      List of BuildCommands representing each command found in the
      presubmit_step traces.
    """
    build_steps: list[BuildCommand] = []

    presubmit_step(ctx)

    step_traces = get_check_traces(ctx)

    for trace in step_traces:
        trace_args = list(trace.args)
        # Check for ninja -t graph command and skip it
        if trace_args[0].endswith('ninja'):
            try:
                dash_t_index = trace_args.index('-t')
                graph_index = trace_args.index('graph')
                if graph_index == dash_t_index + 1:
                    # This trace has -t graph, skip it.
                    continue
            except ValueError:
                # '-t graph' was not found
                pass

        if 'pw_package_install' in trace.call_annotation:
            build_steps.append(_pw_package_install_to_build_command(trace))
            continue

        if 'bazel' in trace.args:
            build_steps.extend(_bazel_command_args_to_build_commands(trace))
            continue

        # Check for gn gen or pw-wrap-ninja
        transformed_args = []
        pw_wrap_ninja_found = False
        gn_found = False
        gn_gen_found = False

        for arg in trace.args:
            # Check for a 'gn gen' command
            if arg == 'gn':
                gn_found = True
            if arg == 'gen' and gn_found:
                gn_gen_found = True

            # Check for pw-wrap-ninja, pw build doesn't use this.
            if arg == 'pw-wrap-ninja':
                # Use ninja instead
                transformed_args.append('ninja')
                pw_wrap_ninja_found = True
                continue
            # Remove --log-actions if pw-wrap-ninja was found. This is a
            # non-standard ninja arg.
            if pw_wrap_ninja_found and arg == '--log-actions':
                continue
            transformed_args.append(str(arg))

        if gn_gen_found:
            # Run the command with run_if=should_gn_gen
            build_steps.append(
                BuildCommand(run_if=should_gn_gen, command=transformed_args)
            )
        else:
            # Run the command as is.
            build_steps.append(BuildCommand(command=transformed_args))

    return build_steps


def presubmit_build_recipe(  # pylint: disable=too-many-locals
    repo_root: Path,
    presubmit_out_dir: Path,
    package_root: Path,
    presubmit_step: Check,
    all_files: list[Path],
    modified_files: list[Path],
) -> BuildRecipe | None:
    """Construct a BuildRecipe from a pw_presubmit step."""
    out_dir = presubmit_out_dir / presubmit_step.name

    ctx = PresubmitContext(
        root=repo_root,
        repos=(repo_root,),
        output_dir=out_dir,
        failure_summary_log=out_dir / 'failure-summary.log',
        paths=tuple(modified_files),
        all_paths=tuple(all_files),
        package_root=package_root,
        luci=None,
        override_gn_args={},
        num_jobs=None,
        continue_after_build_error=True,
        _failed=False,
        format_options=pw_presubmit.presubmit.FormatOptions.load(),
        dry_run=True,
    )

    presubmit_instance = Presubmit(
        root=repo_root,
        repos=(repo_root,),
        output_directory=out_dir,
        paths=modified_files,
        all_paths=all_files,
        package_root=package_root,
        override_gn_args={},
        continue_after_build_error=True,
        rng_seed=1,
        full=False,
    )

    program = Program('', [presubmit_step])
    checks = list(presubmit_instance.apply_filters(program))
    if not checks:
        _LOG.warning('')
        _LOG.warning(
            'Step "%s" is not required for the current set of modified files.',
            presubmit_step.name,
        )
        _LOG.warning('')
        return None

    try:
        ctx.paths = tuple(checks[0].paths)
    except IndexError:
        raise PresubmitTraceAnnotationError(
            'Missing pw_presubmit.presubmit.Check for presubmit step:\n'
            + repr(presubmit_step)
        )

    if isinstance(presubmit_step, GnGenNinja):
        # GnGenNinja is directly translatable to a BuildRecipe.
        selected_gn_args = {
            name: value(ctx) if callable(value) else value
            for name, value in presubmit_step.gn_args.items()
        }

        return BuildRecipe(
            build_dir=out_dir,
            title=presubmit_step.name,
            steps=[
                _pw_package_install_command(name)
                for name in presubmit_step._packages  # pylint: disable=protected-access
            ]
            + [
                BuildCommand(
                    run_if=should_gn_gen,
                    command=[
                        'gn',
                        'gen',
                        str(out_dir),
                        gn_args(**selected_gn_args),
                    ],
                ),
                BuildCommand(
                    build_system_command='ninja',
                    targets=presubmit_step.ninja_targets,
                ),
            ],
        )

    # Unknown type of presubmit, use dry-run to capture subprocess traces.
    build_steps = _presubmit_trace_to_build_commands(ctx, presubmit_step)

    out_dir.mkdir(parents=True, exist_ok=True)

    return BuildRecipe(
        build_dir=out_dir,
        title=presubmit_step.name,
        steps=build_steps,
    )


def get_parser(
    presubmit_programs: Programs | None = None,
    build_recipes: list[BuildRecipe] | None = None,
) -> argparse.ArgumentParser:
    """Setup argparse for pw_build.project_builder and optionally pw_watch."""
    parser = argparse.ArgumentParser(
        prog='pw build',
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    if PW_WATCH_AVAILABLE:
        parser = add_watch_arguments(parser)
    else:
        parser = add_project_builder_arguments(parser)

    if build_recipes is not None:

        def build_recipe_argparse_type(arg: str) -> list[BuildRecipe]:
            """Return a list of matching presubmit steps."""
            assert build_recipes
            all_recipe_names = list(
                recipe.display_name for recipe in build_recipes
            )
            filtered_names = fnmatch.filter(all_recipe_names, arg)

            if not filtered_names:
                recipe_name_str = '\n'.join(sorted(all_recipe_names))
                raise argparse.ArgumentTypeError(
                    f'"{arg}" does not match the name of a recipe.\n\n'
                    f'Valid Recipes:\n{recipe_name_str}'
                )

            return list(
                recipe
                for recipe in build_recipes
                if recipe.display_name in filtered_names
            )

        parser.add_argument(
            '-r',
            '--recipe',
            action='extend',
            default=[],
            help=(
                'Run a build recipe. Include an asterix to match more than one '
                "name. For example: --recipe 'gn_*'"
            ),
            type=build_recipe_argparse_type,
        )

    if presubmit_programs is not None:
        # Add presubmit step arguments.
        all_steps = presubmit_programs.all_steps()

        def presubmit_step_argparse_type(arg: str) -> list[Check]:
            """Return a list of matching presubmit steps."""
            filtered_step_names = fnmatch.filter(all_steps.keys(), arg)

            if not filtered_step_names:
                all_step_names = '\n'.join(sorted(all_steps.keys()))
                raise argparse.ArgumentTypeError(
                    f'"{arg}" does not match the name of a presubmit step.\n\n'
                    f'Valid Steps:\n{all_step_names}'
                )

            return list(all_steps[name] for name in filtered_step_names)

        parser.add_argument(
            '-s',
            '--step',
            action='extend',
            default=[],
            help=(
                'Run presubmit step. Include an asterix to match more than one '
                "step name. For example: --step '*_format'"
            ),
            type=presubmit_step_argparse_type,
        )

    if build_recipes or presubmit_programs:
        parser.add_argument(
            '-l',
            '--list',
            action='store_true',
            default=False,
            help=('List all known build recipes and presubmit steps.'),
        )

    if build_recipes:
        parser.add_argument(
            '--all',
            action='store_true',
            default=False,
            help=('Run all known build recipes.'),
        )

    parser.add_argument(
        '--progress-bars',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Show progress bars in the terminal.',
    )

    parser.add_argument(
        '--log-build-steps',
        action=argparse.BooleanOptionalAction,
        help='Show ninja build step log lines in output.',
    )

    if PW_WATCH_AVAILABLE:
        parser.add_argument(
            '-w',
            '--watch',
            action='store_true',
            help='Use pw_watch to monitor changes.',
            default=False,
        )

    parser.add_argument(
        '-b',
        '--base',
        help=(
            'Git revision to diff for changed files. This is used for '
            'presubmit steps.'
        ),
    )

    parser = add_tab_complete_arguments(parser)

    parser.add_argument(
        '--tab-complete-recipe',
        nargs='?',
        help='Print tab completions for the supplied recipe name.',
    )

    parser.add_argument(
        '--tab-complete-presubmit-step',
        nargs='?',
        help='Print tab completions for the supplied presubmit name.',
    )

    return parser


def _get_prefs(
    args: argparse.Namespace,
) -> ProjectBuilderPrefs | WatchAppPrefs:
    """Load either WatchAppPrefs or ProjectBuilderPrefs.

    Applies the command line args to the correct prefs class.

    Returns:
      A WatchAppPrefs instance if pw_watch is importable, ProjectBuilderPrefs
      otherwise.
    """
    prefs: ProjectBuilderPrefs | WatchAppPrefs
    if PW_WATCH_AVAILABLE:
        prefs = WatchAppPrefs(load_argparse_arguments=add_watch_arguments)
        prefs.apply_command_line_args(args)
    else:
        prefs = ProjectBuilderPrefs(
            load_argparse_arguments=add_project_builder_arguments
        )
        prefs.apply_command_line_args(args)
    return prefs


def load_presubmit_build_recipes(
    presubmit_programs: Programs,
    presubmit_steps: list[Check],
    repo_root: Path,
    presubmit_out_dir: Path,
    package_root: Path,
    all_files: list[Path],
    modified_files: list[Path],
    default_presubmit_step_names: list[str] | None = None,
) -> list[BuildRecipe]:
    """Convert selected presubmit steps into a list of BuildRecipes."""
    # Use the default presubmit if no other steps or command line out
    # directories are provided.
    if len(presubmit_steps) == 0 and default_presubmit_step_names:
        default_steps = list(
            check
            for name, check in presubmit_programs.all_steps().items()
            if name in default_presubmit_step_names
        )
        presubmit_steps = default_steps

    presubmit_recipes: list[BuildRecipe] = []

    for step in presubmit_steps:
        build_recipe = presubmit_build_recipe(
            repo_root,
            presubmit_out_dir,
            package_root,
            step,
            all_files,
            modified_files,
        )
        if build_recipe:
            presubmit_recipes.append(build_recipe)

    return presubmit_recipes


def _tab_complete_recipe(
    build_recipes: list[BuildRecipe],
    text: str = '',
) -> None:
    for name in sorted(recipe.display_name for recipe in build_recipes):
        if name.startswith(text):
            print(name)


def _tab_complete_presubmit_step(
    presubmit_programs: Programs,
    text: str = '',
) -> None:
    for name in sorted(presubmit_programs.all_steps().keys()):
        if name.startswith(text):
            print(name)


def _list_steps_and_recipes(
    presubmit_programs: Programs | None = None,
    build_recipes: list[BuildRecipe] | None = None,
) -> None:
    if presubmit_programs:
        _LOG.info('Presubmit steps:')
        print()
        for name in sorted(presubmit_programs.all_steps().keys()):
            print(name)
        print()
    if build_recipes:
        _LOG.info('Build recipes:')
        print()
        for name in sorted(recipe.display_name for recipe in build_recipes):
            print(name)
        print()


def _print_usage_help(
    presubmit_programs: Programs | None = None,
    build_recipes: list[BuildRecipe] | None = None,
) -> None:
    """Print usage examples with known presubmits and build recipes."""

    def print_pw_build(
        option: str, arg: str | None = None, end: str = '\n'
    ) -> None:
        print(
            ' '.join(
                [
                    'pw build',
                    _COLOR.cyan(option),
                    _COLOR.yellow(arg) if arg else '',
                ]
            ),
            end=end,
        )

    if presubmit_programs:
        print(_COLOR.green('All presubmit steps:'))
        for name in sorted(presubmit_programs.all_steps().keys()):
            print_pw_build('--step', name)
    if build_recipes:
        if presubmit_programs:
            # Add a blank line separator
            print()
        print(_COLOR.green('All build recipes:'))
        for name in sorted(recipe.display_name for recipe in build_recipes):
            print_pw_build('--recipe', name)

        print()
        print(
            _COLOR.green(
                'Recipe and step names may use wildcards and be repeated:'
            )
        )
        print_pw_build('--recipe', '"default_*"', end=' ')
        print(
            _COLOR.cyan('--step'),
            _COLOR.yellow('step1'),
            _COLOR.cyan('--step'),
            _COLOR.yellow('step2'),
        )
        print()
        print(_COLOR.green('Run all build recipes:'))
        print_pw_build('--all')
        print()
        print(_COLOR.green('For more help please run:'))
        print_pw_build('--help')


def main(
    presubmit_programs: Programs | None = None,
    default_presubmit_step_names: list[str] | None = None,
    build_recipes: list[BuildRecipe] | None = None,
    default_build_recipe_names: list[str] | None = None,
    repo_root: Path | None = None,
    presubmit_out_dir: Path | None = None,
    package_root: Path | None = None,
    default_root_logfile: Path = Path('out/build.txt'),
    force_pw_watch: bool = False,
) -> int:
    """Build upstream Pigweed presubmit steps."""
    # pylint: disable=too-many-locals,too-many-branches
    parser = get_parser(presubmit_programs, build_recipes)
    args = parser.parse_args()

    if args.tab_complete_option is not None:
        print_completions_for_option(
            parser,
            text=args.tab_complete_option,
            tab_completion_format=args.tab_complete_format,
        )
        return 0

    log_level = logging.DEBUG if args.debug_logging else logging.INFO

    pw_cli.log.install(
        level=log_level,
        use_color=args.colors,
        # Hide the date from the timestamp
        time_format='%H:%M:%S',
    )

    pw_env = pw_cli.env.pigweed_environment()
    if pw_env.PW_EMOJI:
        charset = EMOJI_CHARSET
    else:
        charset = ASCII_CHARSET

    if args.tab_complete_recipe is not None:
        if build_recipes:
            _tab_complete_recipe(build_recipes, text=args.tab_complete_recipe)
        # Must exit if there are no build_recipes.
        return 0

    if args.tab_complete_presubmit_step is not None:
        if presubmit_programs:
            _tab_complete_presubmit_step(
                presubmit_programs, text=args.tab_complete_presubmit_step
            )
        # Must exit if there are no presubmit_programs.
        return 0

    # List valid steps + recipes.
    if hasattr(args, 'list') and args.list:
        _list_steps_and_recipes(presubmit_programs, build_recipes)
        return 0

    command_line_dash_c_recipes: list[BuildRecipe] = []
    # If -C out directories are provided add them to the recipes list.
    if args.build_directories:
        prefs = _get_prefs(args)
        command_line_dash_c_recipes = create_build_recipes(prefs)

    if repo_root is None:
        repo_root = pw_env.PW_PROJECT_ROOT
    if presubmit_out_dir is None:
        presubmit_out_dir = repo_root / 'out/presubmit'
    if package_root is None:
        package_root = pw_env.PW_PACKAGE_ROOT

    all_files: list[Path]
    modified_files: list[Path]

    all_files, modified_files = fetch_file_lists(
        root=repo_root,
        repo=repo_root,
        pathspecs=[],
        base=args.base,
    )

    # Log modified file summary just like pw_presubmit if using --base.
    if args.base:
        _LOG.info(
            'Running steps that apply to modified files since "%s":', args.base
        )
        _LOG.info('')
        for line in file_summary(
            mf.relative_to(repo_root) for mf in modified_files
        ):
            _LOG.info(line)
        _LOG.info('')

    selected_build_recipes: list[BuildRecipe] = []
    if build_recipes:
        if hasattr(args, 'recipe'):
            selected_build_recipes = args.recipe
        if not selected_build_recipes and default_build_recipe_names:
            selected_build_recipes = [
                recipe
                for recipe in build_recipes
                if recipe.display_name in default_build_recipe_names
            ]

    selected_presubmit_recipes: list[BuildRecipe] = []
    if presubmit_programs and hasattr(args, 'step'):
        selected_presubmit_recipes = load_presubmit_build_recipes(
            presubmit_programs,
            args.step,
            repo_root,
            presubmit_out_dir,
            package_root,
            all_files,
            modified_files,
            default_presubmit_step_names=default_presubmit_step_names,
        )

    # If no builds specifed on the command line print a useful help message:
    if (
        not selected_build_recipes
        and not command_line_dash_c_recipes
        and not selected_presubmit_recipes
        and not args.all
    ):
        _print_usage_help(presubmit_programs, build_recipes)
        return 1

    if build_recipes and args.all:
        selected_build_recipes = build_recipes

    # Run these builds in order:
    recipes_to_build = (
        # -C dirs
        command_line_dash_c_recipes
        # --step 'name'
        + selected_presubmit_recipes
        # --recipe 'name'
        + selected_build_recipes
    )

    # Always set separate build file logging.
    if not args.logfile:
        args.logfile = default_root_logfile
    if not args.separate_logfiles:
        args.separate_logfiles = True

    workers = 1
    if args.parallel:
        # If parallel is requested and parallel_workers is set to 0 run all
        # recipes in parallel. That is, use the number of recipes as the worker
        # count.
        if args.parallel_workers == 0:
            workers = len(recipes_to_build)
        else:
            workers = args.parallel_workers

    project_builder = ProjectBuilder(
        build_recipes=recipes_to_build,
        jobs=args.jobs,
        banners=args.banners,
        keep_going=args.keep_going,
        colors=args.colors,
        charset=charset,
        separate_build_file_logging=args.separate_logfiles,
        # If running builds in serial, send all sub build logs to the root log
        # window (or terminal).
        send_recipe_logs_to_root=(workers == 1),
        root_logfile=args.logfile,
        root_logger=_LOG,
        log_level=log_level,
        allow_progress_bars=args.progress_bars,
        log_build_steps=args.log_build_steps,
    )

    if project_builder.should_use_progress_bars():
        project_builder.use_stdout_proxy()

    if PW_WATCH_AVAILABLE and (
        force_pw_watch or (args.watch or args.fullscreen)
    ):
        event_handler, exclude_list = watch_setup(
            project_builder,
            parallel=args.parallel,
            parallel_workers=workers,
            fullscreen=args.fullscreen,
            logfile=args.logfile,
            separate_logfiles=args.separate_logfiles,
        )

        run_watch(
            event_handler,
            exclude_list,
            fullscreen=args.fullscreen,
        )
        return 0

    # One off build
    return run_builds(project_builder, workers)



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""JSON validity check."""

import json

from . import presubmit, presubmit_context


@presubmit.filter_paths(endswith=('.json',))
@presubmit.check(name='json_check')
def presubmit_check(ctx: presubmit_context.PresubmitContext):
    """Presubmit check that ensures JSON files are valid."""

    ctx.paths = presubmit_context.apply_exclusions(ctx)

    for path in ctx.paths:
        with path.open('r') as ins:
            try:
                json.load(ins)
            except json.decoder.JSONDecodeError as exc:
                intro_line = f'failed to parse {path.relative_to(ctx.root)}'
                with ctx.failure_summary_log.open('w') as outs:
                    print(intro_line, file=outs)
                    print(exc, file=outs)
                ctx.fail(intro_line)
                ctx.fail(str(exc))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_protobuf_protos/field_options.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n&pw_protobuf_protos/field_options.proto\x12\x0bpw.protobuf\x1a google/protobuf/descriptor.proto\"X\n\x0c\x46ieldOptions\x12\x16\n\tmax_count\x18\x01 \x01(\x05H\x00\x88\x01\x01\x12\x15\n\x08max_size\x18\x02 \x01(\x05H\x01\x88\x01\x01\x42\x0c\n\n_max_countB\x0b\n\t_max_size:J\n\x04pwpb\x12\x1d.google.protobuf.FieldOptions\x18\x83\t \x01(\x0b\x32\x19.pw.protobuf.FieldOptions\x88\x01\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'pw_protobuf_protos.field_options_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_FIELDOPTIONS']._serialized_start=89
  _globals['_FIELDOPTIONS']._serialized_end=177
# @@protoc_insertion_point(module_scope)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Wrapers for pyserial classes to log read and write data."""
from __future__ import annotations

from typing import TYPE_CHECKING

import serial

from pw_console.plugins.bandwidth_toolbar import SerialBandwidthTracker

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer


class SerialWithLogging(serial.Serial):  # pylint: disable=too-many-ancestors
    """pyserial with read and write wrappers for logging."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bandwidth_tracker = SerialBandwidthTracker()

    def read(self, size: int = 1) -> bytes:
        data = super().read(size)
        self._bandwidth_tracker.track_read_data(data)
        return data

    def write(self, data: ReadableBuffer) -> None:
        self._bandwidth_tracker.track_write_data(data)
        super().write(data)



# Copyright 2024 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Device instance creation."""

import argparse
from dataclasses import dataclass
import logging
from pathlib import Path
import time
from types import ModuleType
from typing import Any, Callable, Collection

import serial

from pw_console import pyserial_wrapper
from pw_console import socket_client
from pw_hdlc import rpc
from pw_log.log_decoder import timestamp_parser_ms_since_boot
from pw_tokenizer import detokenize

from pw_system.find_serial_port import interactive_serial_port_select
from pw_system import device as pw_device
from pw_system import device_tracing as pw_device_tracing

# Default proto imports:
# pylint: disable=ungrouped-imports,wrong-import-order
from pw_file import file_pb2
from pw_log.proto import log_pb2
from pw_metric_proto import metric_service_pb2
from pw_rpc import echo_pb2
from pw_stream import stream_readers
from pw_thread_protos import thread_snapshot_service_pb2
from pw_trace_protos import trace_service_pb2

from pw_system_protos import device_service_pb2
from pw_unit_test_proto import unit_test_pb2


# pylint: enable=ungrouped-imports

# Internal log for troubleshooting this tool (the console).
_LOG = logging.getLogger(__package__)
_ROOT_LOG = logging.getLogger()

DEFAULT_DEVICE_LOGGER = logging.getLogger('rpc_device')


@dataclass
class DeviceConnection:
    """Stores a Device client along with the reader and writer."""

    client: pw_device_tracing.DeviceWithTracing | pw_device.Device
    reader: stream_readers.SelectableReader | stream_readers.SerialReader
    writer: Callable[[bytes], int | None]

    def __enter__(self):
        """Enter the reader followed by the client context managers.

        Returns the device client for RPC interaction.
        """
        self.reader.__enter__()
        self.client.__enter__()
        return self.client

    def __exit__(self, *exc_info):
        """Close the device connection followed by the reader."""
        self.client.__exit__()
        self.reader.__exit__()


def create_device_serial_or_socket_connection(
    # pylint: disable=too-many-arguments,too-many-locals
    device: str,
    baudrate: int,
    token_databases: Collection[Path],
    socket_addr: str | None = None,
    ticks_per_second: int | None = None,
    serial_debug: bool = False,
    compiled_protos: list[ModuleType] | None = None,
    rpc_logging: bool = True,
    channel_id: int = rpc.DEFAULT_CHANNEL_ID,
    hdlc_encoding: bool = True,
    device_tracing: bool = True,
    device_class: type[pw_device.Device] | None = pw_device.Device,
    device_tracing_class: type[pw_device_tracing.DeviceWithTracing]
    | None = (pw_device_tracing.DeviceWithTracing),
    timestamp_decoder: Callable[[int], str] | None = None,
) -> DeviceConnection:
    """Setup a pw_system.Device client connection.

    Full example of opening a device connection and running an RPC:

    .. code-block:: python

       from pw_system.device_connection import (
           add_device_args,
           create_device_serial_or_socket_connection,
       )

       from pw_protobuf_protos import common_pb2
       from pw_rpc import echo_pb2

       parser = argparse.ArgumentParser(
           prog='MyProductScript',
       )
       parser = add_device_args(parser)
       args = parser.parse_args()

       compiled_protos = [
           common_pb2,
           echo_pb2,
       ]

       device_connection = create_device_serial_or_socket_connection(
           device=args.device,
           baudrate=args.baudrate,
           token_databases=args.token_databases,
           compiled_protos=compiled_protos,
           socket_addr=args.socket_addr,
           ticks_per_second=args.ticks_per_second,
           serial_debug=args.serial_debug,
           rpc_logging=args.rpc_logging,
           hdlc_encoding=args.hdlc_encoding,
           channel_id=args.channel_id,
           device_tracing=args.device_tracing,
           device_class=Device,
           device_tracing_class=DeviceWithTracing,
           timestamp_decoder=timestamp_parser_ms_since_boot,
       )


       # Open the device connction and interact with the Device client.
       with device_connection as device:
           # Make a shortcut to the EchoService.
           echo_service = device.rpcs.pw.rpc.EchoService

           # Call some RPCs and check the results.
           result = echo_service.Echo(msg='Hello')

           if result.status.ok():
               print('The status was', result.status)
               print('The message was', result.response.msg)
           else:
               print('Uh oh, this RPC returned', result.status)

           # The result object can be split into status and payload
           # when assigned.
           status, payload = echo_service.Echo(msg='Goodbye!')

           print(f'{status}: {payload}')
    """

    detokenizer = None
    if token_databases:
        token_databases_with_domains = [] * len(token_databases)
        for token_database in token_databases:
            # Load all domains from token database.
            token_databases_with_domains.append(str(token_database) + "#.*")

        detokenizer = detokenize.AutoUpdatingDetokenizer(
            *token_databases_with_domains
        )
        detokenizer.show_errors = True

    protos: list[ModuleType | Path] = []

    if compiled_protos is None:
        compiled_protos = []

    # Append compiled log.proto library to avoid include errors when
    # manually provided, and shadowing errors due to ordering when the
    # default global search path is used.
    if rpc_logging:
        compiled_protos.append(log_pb2)
    compiled_protos.append(unit_test_pb2)
    protos.extend(compiled_protos)
    protos.append(metric_service_pb2)
    protos.append(thread_snapshot_service_pb2)
    protos.append(file_pb2)
    protos.append(echo_pb2)
    protos.append(trace_service_pb2)
    protos.append(device_service_pb2)

    reader: stream_readers.SelectableReader | stream_readers.SerialReader

    if socket_addr is None:
        serial_impl = (
            pyserial_wrapper.SerialWithLogging
            if serial_debug
            else serial.Serial
        )

        if not device:
            device = interactive_serial_port_select()
        _ROOT_LOG.info('Using serial port: %s', device)
        serial_device = serial_impl(
            device,
            baudrate,
            # Timeout in seconds. This should be a very small value. Setting to
            # zero makes pyserial read() non-blocking which will cause the host
            # machine to busy loop and 100% CPU usage.
            # https://pythonhosted.org/pyserial/pyserial_api.html#serial.Serial
            timeout=0.1,
        )
        reader = stream_readers.SerialReader(serial_device, 8192)
        write = serial_device.write

        # Overwrite decoder for serial device.
        if timestamp_decoder is None:
            timestamp_decoder = timestamp_parser_ms_since_boot
    else:
        socket_impl = (
            socket_client.SocketClientWithLogging
            if serial_debug
            else socket_client.SocketClient
        )

        def disconnect_handler(
            socket_device: socket_client.SocketClient,
        ) -> None:
            """Attempts to reconnect on disconnected socket."""
            _LOG.error('Socket disconnected. Will retry to connect.')
            while True:
                try:
                    socket_device.connect()
                    break
                except:  # pylint: disable=bare-except
                    # Ignore errors and retry to reconnect.
                    time.sleep(1)
            _LOG.info('Successfully reconnected')

        try:
            socket_device = socket_impl(
                socket_addr, on_disconnect=disconnect_handler
            )
            reader = stream_readers.SelectableReader(socket_device)
            write = socket_device.write
        except ValueError as error:
            raise ValueError(
                f'Failed to initialize socket at {socket_addr}'
            ) from error

    device_args: list[Any] = [channel_id, reader, write]
    device_kwds: dict[str, Any] = {
        'proto_library': protos,
        'detokenizer': detokenizer,
        'timestamp_decoder': timestamp_decoder,
        'rpc_timeout_s': 5,
        'use_rpc_logging': rpc_logging,
        'use_hdlc_encoding': hdlc_encoding,
    }

    device_client: pw_device_tracing.DeviceWithTracing | pw_device.Device
    if device_tracing:
        device_kwds['ticks_per_second'] = ticks_per_second
        if device_tracing_class is None:
            device_tracing_class = pw_device_tracing.DeviceWithTracing
        device_client = device_tracing_class(*device_args, **device_kwds)
    else:
        if device_class is None:
            device_class = pw_device.Device
        device_client = device_class(*device_args, **device_kwds)

    return DeviceConnection(device_client, reader, write)


def add_device_args(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Add device specific args required by the pw_system.Device class."""

    group = parser.add_mutually_exclusive_group(required=False)

    group.add_argument(
        '-d',
        '--device',
        help='the serial port to use',
    )
    group.add_argument(
        '-s',
        '--socket-addr',
        type=str,
        help=(
            'Socket address used to connect to server. Type "default" to use '
            'localhost:33000, pass the server address and port as '
            'address:port, or prefix the path to a forwarded socket with '
            f'"{socket_client.SocketClient.FILE_SOCKET_SERVER}:" as '
            f'{socket_client.SocketClient.FILE_SOCKET_SERVER}:path_to_file.'
        ),
    )

    parser.add_argument(
        '-b',
        '--baudrate',
        type=int,
        default=115200,
        help='the baud rate to use',
    )
    parser.add_argument(
        '--serial-debug',
        action='store_true',
        help=(
            'Enable debug log tracing of all data passed through'
            'pyserial read and write.'
        ),
    )
    parser.add_argument(
        "--token-databases",
        metavar='elf_or_token_database',
        nargs="+",
        type=Path,
        help="Path to tokenizer database csv file(s).",
    )
    parser.add_argument(
        '-f',
        '--ticks_per_second',
        type=int,
        dest='ticks_per_second',
        help=('The clock rate of the trace events.'),
    )
    parser.add_argument(
        '--rpc-logging',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Use pw_rpc based logging.',
    )
    parser.add_argument(
        '--hdlc-encoding',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Use HDLC encoding on transfer interfaces.',
    )
    parser.add_argument(
        '--channel-id',
        type=int,
        default=rpc.DEFAULT_CHANNEL_ID,
        help="Channel ID used in RPC communications.",
    )
    parser.add_argument(
        '--device-tracing',
        action=argparse.BooleanOptionalAction,
        default=True,
        help='Use device tracing.',
    )

    return parser



# Copyright 2023 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed qemu frontend."""


import io
import json
import logging
import os
import re
import socket
import sys

from pathlib import Path
from typing import Any

from pw_emu.core import (
    ConfigError,
    Connector,
    Launcher,
    Error,
    InvalidChannelType,
    RunError,
    WrongEmulator,
)

_QMP_LOG = logging.getLogger('pw_qemu.qemu.qmp')


class QmpError(Error):
    """Exception for QMP errors."""

    def __init__(self, err: str):
        super().__init__(err)


class QmpClient:
    """Send qmp requests the server."""

    def __init__(self, stream: io.RawIOBase):
        self._stream = stream

        # Perform the QMP "capabilities negotiation" handshake.
        # https://wiki.qemu.org/Documentation/QMP#Capabilities_Negotiation
        #
        # When the QMP connection is established, QEMU first sends a greeting
        # message with its version and capabilities. Then the client sends
        # 'qmp_capabilities' to exit capabilities negotiation mode. The result
        # is an empty 'return'.
        #
        # self.request() will consume both the initial greeting and the
        # subsequent 'return' response.
        self.request('qmp_capabilities')

    def request(self, cmd: str, args: dict[str, Any] | None = None) -> Any:
        """Issue a command using the qmp interface.

        Returns a map with the response or None if there is no
        response for this command.

        """

        req: dict[str, Any] = {'execute': cmd}
        if args:
            req['arguments'] = args
        _QMP_LOG.debug(' -> {json.dumps(cmd)}')
        self._stream.write(json.dumps(req).encode('utf-8'))
        while True:
            line = self._stream.readline()
            _QMP_LOG.debug(' <- {line}')
            resp = json.loads(line)
            if 'error' in resp.keys():
                raise QmpError(resp['error']['desc'])
            if 'return' in resp.keys():
                return resp['return']


class QemuLauncher(Launcher):
    """Start a new qemu process for a given target and config file."""

    def __init__(self, config_path: Path | None = None):
        super().__init__('qemu', config_path)
        self._start_cmd: list[str] = []
        self._chardevs_id_to_name = {
            'compat_monitor0': 'qmp',
            'compat_monitor1': 'monitor',
            'gdb': 'gdb',
        }
        self._chardevs: dict[str, Any] = {}
        self._qmp_init_sock: socket.socket | None = None

    def _set_qemu_channel_tcp(self, name: str, filename: str) -> None:
        """Parse a TCP chardev and return (host, port) tuple.

        Format for the tcp chardev backend:

        [disconnected|isconnected:]tcp:<host>:<port>[,<options>][ <->
        <host>:<port>]

        """

        host_port: Any = filename.split(',')[0]
        if host_port.split(':')[0] != 'tcp':
            host_port = host_port.split(':')[2:]
        else:
            host_port = host_port.split(':')[1:]
        # IPV6 hosts have :
        host = ':'.join(host_port[0:-1])
        port = host_port[-1]
        self._handles.add_channel_tcp(name, host, int(port))

    def _set_qemu_channel_pty(self, name: str, filename: str) -> None:
        """Parse a PTY chardev and return the path.

        Format for the pty chardev backend: pty:<path>
        """

        path = filename.split(':')[1]

        self._handles.add_channel_pty(name, path)

        if os.path.lexists(self._path(name)):
            os.unlink(self._path(name))
        os.symlink(path, self._path(name))

    def _set_qemu_channel(self, name: str, filename: str) -> None:
        """Setups a chardev channel type."""

        if filename.startswith('pty'):
            self._set_qemu_channel_pty(name, filename)
        elif 'tcp' in filename:
            self._set_qemu_channel_tcp(name, filename)

    def _get_channels_config(self, chan: str, opt: str) -> Any:
        val = self._config.get_emu(['channels', chan, opt])
        if val is not None:
            return val
        return self._config.get_emu(['channels', opt])

    def _configure_default_channels(self) -> None:
        """Configure the default channels."""

        # keep qmp first so that it gets the compat_monitor0 label
        for chan in ['qmp', 'monitor', 'gdb']:
            chan_type = self._get_channels_config(chan, 'type')
            if not chan_type:
                chan_type = 'tcp'
            if chan_type == 'pty':
                if sys.platform == 'win32':
                    raise InvalidChannelType(chan_type)
                backend = 'pty'
            elif chan_type == 'tcp':
                backend = 'tcp:localhost:0,ipv4=on,server=on,wait=off'
            else:
                raise InvalidChannelType(chan_type)
            self._start_cmd.extend([f'-{chan}', backend])

    def _get_chardev_config(self, name: str, opt: str) -> Any:
        val = self._config.get_target_emu(['channels', 'chardevs', name, opt])
        if not val:
            val = self._get_channels_config(name, opt)
        return val

    def _configure_serial_channels(self, serials: dict) -> None:
        """Create "standard" serial devices.

        We can't control the serial allocation number for "standard"
        -serial devices so fill the slots for the not needed serials
        with null chardevs e.g. for serial3, serial1 generate the
        following arguments, in this order:

         -serial null -serial {backend} -serial null - serial {backend}

        """

        min_ser = sys.maxsize
        max_ser = -1
        for serial in serials.keys():
            num = int(serial.split('serial')[1])
            if num < min_ser:
                min_ser = num
            if num > max_ser:
                max_ser = num
        for i in range(min_ser, max_ser + 1):
            if serials.get(f'serial{i}'):
                name = serials[f'serial{i}']
                chan_type = self._get_chardev_config(name, 'type')
                if not chan_type:
                    chan_type = 'tcp'
                if chan_type == 'pty':
                    backend = 'pty'
                elif chan_type == 'tcp':
                    backend = 'tcp:localhost:0,ipv4=on,server=on,wait=off'
                else:
                    raise InvalidChannelType(chan_type)
                self._start_cmd.extend(['-serial', backend])
            else:
                self._start_cmd.extend(['-serial', 'null'])

    def _configure_chardev_channels(self) -> None:
        """Configure chardevs."""

        self._chardevs = self._config.get_target_emu(
            ['channels', 'chardevs'], True, dict
        )

        serials = {}
        for name, config in self._chardevs.items():
            chardev_id = config['id']
            self._chardevs_id_to_name[chardev_id] = name

            chardev_type = self._get_chardev_config(name, 'type')
            if chardev_type is None:
                chardev_type = 'tcp'

            if chardev_type == 'pty':
                backend = 'pty'
            elif chardev_type == 'tcp':
                backend = 'socket,host=localhost,port=0,server=on,wait=off'
            else:
                raise InvalidChannelType(chardev_type)

            # serials are configured differently
            if re.search(r'serial[0-9]*', chardev_id):
                serials[chardev_id] = name
            else:
                self._start_cmd.extend(
                    ['-chardev', f'{backend},id={chardev_id}']
                )

        self._configure_serial_channels(serials)

    def _pre_start(
        self,
        target: str,
        file: Path | None = None,
        pause: bool = False,
        debug: bool = False,
        args: str | None = None,
    ) -> list[str]:
        qemu = self._config.get_target_emu(['executable'])
        if not qemu:
            qemu = self._config.get_emu(['executable'], optional=False)
        machine = self._config.get_target_emu(['machine'], optional=False)

        self._start_cmd = [f'{qemu}', '-nographic', '-nodefaults']
        self._start_cmd.extend(['-display', 'none'])
        self._start_cmd.extend(['-machine', f'{machine}'])

        try:
            self._configure_default_channels()
            self._configure_chardev_channels()
        except KeyError as err:
            raise ConfigError(self._config.path, str(err))

        if pause:
            self._start_cmd.append('-S')
        if debug:
            self._start_cmd.extend(['-d', 'guest_errors'])

        if file:
            self._start_cmd.extend(['-kernel', str(file)])

        self._start_cmd.extend(self._config.get_emu(['args'], entry_type=list))
        self._start_cmd.extend(
            self._config.get_target_emu(['args'], entry_type=list)
        )
        if args:
            self._start_cmd.extend(args.split(' '))

        # initial/bootstrap qmp connection
        self._qmp_init_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._qmp_init_sock.bind(('localhost', 0))
        port = self._qmp_init_sock.getsockname()[1]
        self._qmp_init_sock.listen()
        self._qmp_init_sock.settimeout(30)
        self._start_cmd.extend(['-qmp', f'tcp:localhost:{port}'])

        return self._start_cmd

    def _post_start(self) -> None:
        assert self._qmp_init_sock is not None
        try:
            conn, _ = self._qmp_init_sock.accept()
        except (KeyboardInterrupt, socket.timeout):
            raise RunError('qemu', 'qmp connection failed')
        self._qmp_init_sock.close()
        try:
            qmp = QmpClient(conn.makefile('rwb', buffering=0))
        except json.decoder.JSONDecodeError:
            raise RunError('qemu', 'qmp handshake failed')
        conn.close()

        resp = qmp.request('query-chardev')
        for chardev in resp:
            label = chardev['label']
            name = self._chardevs_id_to_name.get(label)
            if name:
                self._set_qemu_channel(name, chardev['filename'])

    def _get_connector(self, wdir: Path) -> Connector:
        return QemuConnector(wdir)


class QemuConnector(Connector):
    """qemu implementation for the emulator specific connector methods."""

    def __init__(self, wdir: Path) -> None:
        super().__init__(wdir)
        if self.get_emu() != 'qemu':
            raise WrongEmulator('qemu', self.get_emu())
        self._qmp: QmpClient | None = None

    def _q(self) -> QmpClient:
        if not self._qmp:
            self._qmp = QmpClient(self.get_channel_stream('qmp'))
        return self._qmp

    def reset(self) -> None:
        self._q().request('system_reset')

    def cont(self) -> None:
        self._q().request('cont')

    def set_property(self, path: str, prop: str, value: Any) -> None:
        args = {
            'path': '{}'.format(path),
            'property': prop,
            'value': value,
        }
        self._q().request('qom-set', args)

    def get_property(self, path: str, prop: str) -> Any:
        args = {
            'path': '{}'.format(path),
            'property': prop,
        }
        return self._q().request('qom-get', args)

    def list_properties(self, path: str) -> list[Any]:
        args = {
            'path': '{}'.format(path),
        }
        return self._q().request('qom-list', args)
